# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import pickle
import re

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import torch

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import given, settings
from pytest import mark, raises, warns

from lenskit.data import Dataset, ItemList
from lenskit.data.vocab import Vocabulary
from lenskit.diagnostics import DataWarning
from lenskit.testing import integer_ids

ITEMS = ["a", "b", "c", "d", "e"]
VOCAB = Vocabulary(ITEMS)


def test_empty():
    il = ItemList()

    assert len(il) == 0
    assert il.numbers().shape == (0,)
    assert il.ids().shape == (0,)
    assert il.scores() is None


def test_empty_ids():
    il = ItemList(item_ids=[])
    assert len(il) == 0
    assert il.ids().shape == (0,)
    assert il.scores() is None


def test_empty_nums():
    il = ItemList(item_nums=[])
    assert len(il) == 0
    assert il.numbers().shape == (0,)
    assert il.scores() is None


def test_empty_list():
    il = ItemList([])
    assert len(il) == 0
    assert il.ids().shape == (0,)
    assert il.scores() is None


def test_item_list():
    il = ItemList(item_ids=["one", "two"])

    assert len(il) == 2
    assert il.ids().shape == (2,)

    with raises(RuntimeError, match="item numbers not available"):
        il.numbers()


def test_item_list_ctor():
    il = ItemList(["one", "two"])

    assert len(il) == 2
    assert il.ids().shape == (2,)

    with raises(RuntimeError, match="item numbers not available"):
        il.numbers()


def test_item_list_alias():
    il = ItemList(item_id=["one", "two"])

    assert len(il) == 2
    assert il.ids().shape == (2,)

    with raises(RuntimeError, match="item numbers not available"):
        il.numbers()


def test_item_list_bad_type():
    with raises(TypeError):
        ItemList(item_id=[3.4, 7.2])


def test_item_list_bad_dimension():
    with raises(TypeError):
        ItemList(item_id=[["one", "two"]])


def test_item_num_array():
    il = ItemList(item_nums=np.arange(5))

    assert len(il) == 5
    assert il.numbers().shape == (5,)

    with raises(RuntimeError, match="item IDs not available"):
        il.ids()


def test_item_num_arrow():
    il = ItemList(item_nums=pa.array(np.arange(5)))

    assert len(il) == 5
    assert il.numbers().shape == (5,)
    assert np.all(il.numbers() == np.arange(5))

    with raises(RuntimeError, match="item IDs not available"):
        il.ids()


def test_item_num_alias():
    il = ItemList(item_num=np.arange(5))

    assert len(il) == 5
    assert il.numbers().shape == (5,)

    with raises(RuntimeError, match="item IDs not available"):
        il.ids()


def test_item_num_bad_type():
    with raises(TypeError):
        ItemList(item_num=np.random.randn(5))


def test_item_num_bad_dims():
    with raises(TypeError):
        ItemList(item_num=[[1, 3, 8, 4]])


def test_item_ids_num_mismatch_sizes():
    with raises(TypeError, match="has incorrect shape"):
        ItemList(item_ids=ITEMS, item_num=np.arange(4))


def test_item_num_list_vocab():
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB)

    assert len(il) == 5
    assert il.numbers().shape == (5,)
    assert il.ids().shape == (5,)

    assert all(il.numbers() == np.arange(5))
    assert all(il.ids() == ITEMS)


def test_item_id_list_vocab():
    il = ItemList(item_ids=ITEMS, vocabulary=VOCAB)

    assert len(il) == 5
    assert il.numbers().shape == (5,)
    assert il.ids().shape == (5,)

    assert all(il.numbers() == np.arange(5))
    assert all(il.ids() == ITEMS)


def test_scores():
    data = np.random.randn(5).astype(np.float32)
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data)

    scores = il.scores()
    assert scores is not None
    assert scores.shape == (5,)
    assert np.all(scores == data.astype(np.float32))
    assert scores.dtype == np.float32

    st = il.scores("torch")
    assert isinstance(st, torch.Tensor)
    assert st.shape == (5,)
    assert np.all(st.numpy() == data)

    assert il.ranks() is None


def test_rank_field():
    data = np.random.randn(5).astype(np.float32)
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data, rank=np.arange(5) + 1)
    assert il.ordered
    assert np.all(il.ranks() == np.arange(5) + 1)


def test_torch_rank_field():
    data = np.random.randn(5).astype(np.float32)
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data, rank=torch.arange(5) + 1)
    assert il.ordered
    assert np.all(il.ranks() == np.arange(5) + 1)


def test_ordered_mismatch():
    data = np.random.randn(5).astype(np.float32)
    with warns(DataWarning, match="ordered=False"):
        il = ItemList(
            item_nums=np.arange(5),
            ordered=False,
            vocabulary=VOCAB,
            scores=data,
            rank=np.arange(5) + 1,
        )
    assert not il.ordered
    assert il.ranks() is None


def test_rank_not_one():
    data = np.random.randn(5).astype(np.float32)
    with warns(DataWarning, match="begin with 1"):
        il = ItemList(
            item_nums=np.arange(5), vocabulary=VOCAB, scores=data, rank=np.arange(2, 11, 2)
        )
    assert il.ordered
    assert np.all(il.ranks() == np.arange(2, 11, 2))


def test_scores_pandas_no_index():
    data = np.random.randn(5).astype(np.float32)
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data)

    scores = il.scores("pandas")
    assert scores is not None
    assert scores.shape == (5,)
    assert np.all(scores == data)


def test_scores_pandas_id_index():
    data = np.random.randn(5).astype(np.float32)
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data)
    scores = il.scores("pandas", index="ids")
    assert scores is not None
    assert scores.shape == (5,)
    assert np.all(scores == data)
    assert scores.index.name == "item_id"
    assert np.all(scores.index.values == ITEMS)


def test_scores_pandas_num_index():
    data = np.random.randn(5).astype(np.float32)
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data)
    scores = il.scores("pandas", index="numbers")
    assert scores is not None
    assert scores.shape == (5,)
    assert np.all(scores == data)
    assert scores.index.name == "item_num"
    assert np.all(scores.index.values == np.arange(5))


def test_ranks():
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, ordered=True)
    assert il.ordered

    ranks = il.ranks()
    assert ranks is not None
    assert ranks.shape == (5,)
    assert np.all(ranks == np.arange(1, 6))


def test_numbers_alt_vocab():
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB)

    av = Vocabulary(["A", "B"] + ITEMS)
    nums = il.numbers(vocabulary=av)
    assert np.all(nums == np.arange(2, 7))


@mark.parametrize("src", ["ids", "numbers"])
def test_numbers_numpy(src):
    match src:
        case "ids":
            il = ItemList(item_ids=ITEMS, vocabulary=VOCAB)
        case "numbers":
            il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB)

    nums = il.numbers(format="numpy")
    assert isinstance(nums, np.ndarray)
    assert np.all(nums == np.arange(5))

    nums = il.numbers(format="numpy", vocabulary=Vocabulary(ITEMS))
    assert isinstance(nums, np.ndarray)
    assert np.all(nums == np.arange(5))


@mark.parametrize("src", ["ids", "numbers"])
def test_numbers_torch(src):
    match src:
        case "ids":
            il = ItemList(item_ids=ITEMS, vocabulary=VOCAB)
        case "numbers":
            il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB)

    nums = il.numbers(format="torch")
    assert torch.is_tensor(nums)
    assert torch.all(nums == torch.arange(5))

    nums = il.numbers(format="torch", vocabulary=Vocabulary(ITEMS))
    assert torch.is_tensor(nums)
    assert torch.all(nums == torch.arange(5))


@mark.parametrize("src", ["ids", "numbers"])
def test_numbers_arrow(src):
    match src:
        case "ids":
            il = ItemList(item_ids=ITEMS, vocabulary=VOCAB)
        case "numbers":
            il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB)

    nums = il.numbers(format="arrow")
    assert isinstance(nums, pa.Int32Array)
    assert np.all(nums.to_numpy() == np.arange(5))

    nums = il.numbers(format="arrow", vocabulary=Vocabulary(ITEMS))
    assert isinstance(nums, pa.Int32Array)
    assert np.all(nums.to_numpy() == np.arange(5))


def test_pandas_df():
    data = np.random.randn(5).astype(np.float32)
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data)

    df = il.to_df()
    assert np.all(df["item_id"] == ITEMS)
    assert np.all(df["item_num"] == np.arange(5))
    assert np.all(df["score"] == data)
    assert "rank" not in df.columns


def test_pandas_df_ordered():
    data = np.random.randn(5).astype(np.float32)
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data, ordered=True)

    df = il.to_df()
    assert np.all(df["item_id"] == ITEMS)
    assert np.all(df["item_num"] == np.arange(5))
    assert np.all(df["score"] == data)
    assert np.all(df["rank"] == np.arange(1, 6))


def test_pandas_df_no_numbers():
    data = np.random.randn(5).astype(np.float32)
    il = ItemList(item_ids=ITEMS, vocabulary=VOCAB, scores=data, ordered=True)
    df = il.to_df(numbers=False)
    assert "item_id" in df.columns

    # even with a vocabulary, we should have no numbers
    assert "item_num" not in df.columns


def test_arrow_types(rng: np.random.Generator):
    data = rng.standard_exponential(5).astype(np.float32)
    il = ItemList(item_ids=ITEMS, vocabulary=VOCAB, scores=data)

    assert il.arrow_types() == {"item_id": pa.string(), "score": pa.float32()}
    assert il.arrow_types(ids=False, numbers=True) == {
        "item_num": pa.int32(),
        "score": pa.float32(),
    }
    assert il.arrow_types(numbers=True) == {
        "item_id": pa.string(),
        "item_num": pa.int32(),
        "score": pa.float32(),
    }

    il = ItemList(item_ids=ITEMS, vocabulary=VOCAB, scores=data, ordered=True)
    assert il.arrow_types() == {"item_id": pa.string(), "rank": pa.int32(), "score": pa.float32()}


def test_arrow_table():
    data = np.random.randn(5).astype(np.float32)
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data)

    tbl = il.to_arrow(numbers=True)
    assert isinstance(tbl, pa.Table)
    assert tbl.num_columns == 3
    assert np.all(tbl.column("item_id").to_numpy() == ITEMS)
    assert np.all(tbl.column("item_num").to_numpy() == np.arange(5))
    assert np.all(tbl.column("score").to_numpy() == data)

    tbl = il.to_arrow()
    assert isinstance(tbl, pa.Table)
    assert tbl.num_columns == 2
    assert np.all(tbl.column("item_id").to_numpy() == ITEMS)
    assert np.all(tbl.column("score").to_numpy() == data)

    tbl = il.to_arrow(ids=False, numbers=True)
    assert isinstance(tbl, pa.Table)
    assert tbl.num_columns == 2
    assert np.all(tbl.column("item_num").to_numpy() == np.arange(5))
    assert np.all(tbl.column("score").to_numpy() == data)


def test_arrow_array():
    data = np.random.randn(5).astype(np.float32)
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data)

    tbl = il.to_arrow(numbers=True, type="array")
    assert isinstance(tbl, pa.StructArray)
    tbl = pa.Table.from_struct_array(tbl)
    assert tbl.num_columns == 3
    assert np.all(tbl.column("item_id").to_numpy() == ITEMS)
    assert np.all(tbl.column("item_num").to_numpy() == np.arange(5))
    assert np.all(tbl.column("score").to_numpy() == data)


def test_empty_arrow():
    il = ItemList()

    arr = il.to_arrow()
    assert arr.num_rows == 0


def test_item_list_pickle_compact(ml_ds: Dataset):
    nums = [1, 0, 308, 24, 72]
    il = ItemList(item_nums=nums, vocabulary=ml_ds.items)
    assert len(il) == 5
    assert np.all(il.ids() == ml_ds.items.ids(nums))

    # check that pickling isn't very big (we don't pickle the vocabulary)
    data = pickle.dumps(il)
    print(len(data))
    assert len(data) <= 500

    il2 = pickle.loads(data)
    assert len(il2) == len(il)
    assert np.all(il2.ids() == il.ids())
    assert np.all(il2.numbers() == il.numbers())


def test_item_list_pickle_fields(ml_ds: Dataset):
    row = ml_ds.user_row(user_num=400)
    assert row is not None
    data = pickle.dumps(row)
    r2 = pickle.loads(data)

    assert len(r2) == len(row)
    assert np.all(r2.ids() == row.ids())
    assert np.all(r2.numbers() == row.numbers())
    assert r2.field("rating") is not None
    assert np.all(r2.field("rating") == row.field("rating"))
    assert r2.field("timestamp") is not None
    assert np.all(r2.field("timestamp") == row.field("timestamp"))


def test_subset_mask(ml_ds: Dataset):
    row = ml_ds.user_row(user_num=400)
    assert row is not None
    ratings = row.field("rating")
    assert ratings is not None

    mask = ratings > 3.0
    pos = row[mask]

    assert len(pos) == np.sum(mask)
    assert np.all(pos.ids() == row.ids()[mask])
    assert np.all(pos.numbers() == row.numbers()[mask])
    rf = row.field("rating")
    assert rf is not None
    prf = pos.field("rating")
    assert prf is not None
    assert np.all(prf == rf[mask])
    assert np.all(prf > 3.0)


def test_subset_idx_list(ml_ds: Dataset):
    row = ml_ds.user_row(user_num=400)
    assert row is not None
    ratings = row.field("rating")
    assert ratings is not None

    ks = [0, 5, 15]
    pos = row[ks]

    assert len(pos) == 3
    assert np.all(pos.ids() == row.ids()[ks])
    assert np.all(pos.numbers() == row.numbers()[ks])
    rf = row.field("rating")
    assert rf is not None
    assert np.all(pos.field("rating") == rf[ks])


def test_subset_idx_array(ml_ds: Dataset):
    row = ml_ds.user_row(user_num=400)
    assert row is not None
    ratings = row.field("rating")
    assert ratings is not None

    ks = np.array([0, 5, 15])
    pos = row[ks]

    assert len(pos) == 3
    assert np.all(pos.ids() == row.ids()[ks])
    assert np.all(pos.numbers() == row.numbers()[ks])
    rf = row.field("rating")
    assert rf is not None
    assert np.all(pos.field("rating") == rf[ks])


@given(st.data())
def test_subset_index_array_stress(ml_ds_unchecked: Dataset, data: st.DataObject):
    ml_ds = ml_ds_unchecked
    uno = data.draw(st.integers(0, ml_ds.user_count - 1))
    row = ml_ds.user_row(user_num=uno)
    assert row is not None
    ratings = row.field("rating")
    assert ratings is not None

    ks = data.draw(
        nph.arrays(np.int32, st.integers(0, 1000), elements=st.integers(0, len(row) - 1))
    )
    found = row[ks]

    assert len(found) == len(ks)
    assert np.all(found.ids() == row.ids()[ks])
    assert np.all(found.numbers() == row.numbers()[ks])
    rf = row.field("rating")
    assert rf is not None
    assert np.all(found.field("rating") == rf[ks])


def test_subset_slice(ml_ds: Dataset):
    row = ml_ds.user_row(user_num=400)
    assert row is not None
    ratings = row.field("rating")
    assert ratings is not None

    pos = row[5:10]

    assert len(pos) == 5
    assert np.all(pos.ids() == row.ids()[5:10])
    assert np.all(pos.numbers() == row.numbers()[5:10])
    rf = row.field("rating")
    assert rf is not None
    assert np.all(pos.field("rating") == rf[5:10])


def test_from_df():
    df = pd.DataFrame(
        {"item_id": ITEMS, "item_num": np.arange(5), "score": np.random.randn(5).astype(np.float32)}
    )
    il = ItemList.from_df(df, vocabulary=VOCAB)  # type: ignore
    assert len(il) == 5
    assert np.all(il.ids() == ITEMS)
    assert np.all(il.numbers() == np.arange(5))
    assert np.all(il.scores() == df["score"].values)
    assert not il.ordered


def test_from_df_user():
    df = pd.DataFrame(
        {
            "user_id": 50,
            "item_id": ITEMS,
            "item_num": np.arange(5),
            "score": np.random.randn(5).astype(np.float32),
        }
    )
    il = ItemList.from_df(df, vocabulary=VOCAB)  # type: ignore
    assert len(il) == 5
    assert np.all(il.ids() == ITEMS)
    assert np.all(il.numbers() == np.arange(5))
    assert np.all(il.scores() == df["score"].values)
    assert il.field("user_id") is None


def test_from_df_ranked():
    df = pd.DataFrame(
        {
            "item_id": ITEMS,
            "item_num": np.arange(5),
            "rank": np.arange(1, 6),
            "score": np.random.randn(5).astype(np.float32),
        }
    )
    il = ItemList.from_df(df, vocabulary=VOCAB)  # type: ignore
    assert len(il) == 5
    assert np.all(il.ids() == ITEMS)
    assert np.all(il.numbers() == np.arange(5))
    assert np.all(il.scores() == df["score"].values)
    assert il.ordered
    assert np.all(il.ranks() == np.arange(1, 6))


def test_from_arrow_table():
    df = pd.DataFrame(
        {"item_id": ITEMS, "item_num": np.arange(5), "score": np.random.randn(5).astype(np.float32)}
    )
    arr = pa.Table.from_pandas(df)
    il = ItemList.from_arrow(arr, vocabulary=VOCAB)  # type: ignore
    assert len(il) == 5
    assert np.all(il.ids() == ITEMS)
    assert np.all(il.numbers() == np.arange(5))
    assert np.all(il.scores() == df["score"].values)
    assert not il.ordered


def test_from_arrow_table_nullable():
    df = pd.DataFrame(
        {
            "item_id": np.arange(1, 100),
            "item_num": np.arange(99),
            "score": np.random.randn(99).astype(np.float32),
        }
    )
    df.loc[df["score"] < 0, "score"] = np.nan
    arr = pa.Table.from_pandas(df)
    il = ItemList.from_arrow(arr)  # type: ignore
    assert len(il) == 99
    assert np.all(il.ids() == np.arange(1, 100))
    scores = il.scores()
    assert scores is not None
    assert np.all(np.isfinite(scores) == df["score"].notnull())
    assert np.all(scores[df["score"].notnull()] == df.loc[df["score"].notnull(), "score"])
    assert not il.ordered


def test_from_empty_arrow():
    tbl = pa.table(
        {"item_num": pa.array([], type=pa.int32()), "score": pa.array([], type=pa.float32())}
    )
    print(tbl)
    il = ItemList.from_arrow(tbl, vocabulary=VOCAB)
    assert len(il) == 0
    scores = il.scores()
    assert scores is not None
    assert len(scores) == 0


def test_from_arrow_array():
    df = pd.DataFrame(
        {"item_id": ITEMS, "item_num": np.arange(5), "score": np.random.randn(5).astype(np.float32)}
    )
    arr = pa.Table.from_pandas(df).to_struct_array()
    il = ItemList.from_arrow(arr, vocabulary=VOCAB)  # type: ignore
    assert len(il) == 5
    assert np.all(il.ids() == ITEMS)
    assert np.all(il.numbers() == np.arange(5))
    assert np.all(il.scores() == df["score"].values)
    assert not il.ordered


def test_copy_ctor():
    data = np.random.randn(5).astype(np.float32)
    extra = np.random.randn(5).astype(np.float32)

    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data, extra=extra)
    copy = ItemList(il)
    assert copy is not il
    assert len(copy) == len(il)

    assert np.all(copy.ids() == il.ids())
    assert np.all(copy.numbers() == il.numbers())
    assert np.all(copy.scores() == data)
    assert copy._vocab is il._vocab

    x = copy.field("extra")
    assert x is not None
    assert np.all(x == extra)


def test_copy_na_scores():
    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB)
    il2 = ItemList(il, scores=np.nan)

    scores = il2.scores()
    assert scores is not None
    assert np.all(np.isnan(scores))


def test_copy_ctor_remove_scores():
    data = np.random.randn(5).astype(np.float32)
    extra = np.random.randn(5).astype(np.float32)

    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data, extra=extra)
    copy = ItemList(il, scores=False)
    assert copy is not il
    assert len(copy) == len(il)

    assert np.all(copy.ids() == il.ids())
    assert np.all(copy.numbers() == il.numbers())
    assert copy.scores() is None
    assert copy._vocab is il._vocab

    x = copy.field("extra")
    assert x is not None
    assert np.all(x == extra)


def test_copy_ctor_remove_extra():
    data = np.random.randn(5).astype(np.float32)
    extra = np.random.randn(5).astype(np.float32)

    il = ItemList(item_nums=np.arange(5), vocabulary=VOCAB, scores=data, extra=extra)
    copy = ItemList(il, extra=False)
    assert copy is not il
    assert len(copy) == len(il)

    assert np.all(copy.ids() == il.ids())
    assert np.all(copy.numbers() == il.numbers())
    assert np.all(copy.scores() == data)
    assert copy._vocab is il._vocab

    x = copy.field("extra")
    assert x is None


def test_from_vocab(ml_ds: Dataset):
    items = ItemList.from_vocabulary(ml_ds.items)

    assert len(items) == ml_ds.item_count
    assert np.all(items.ids() == ml_ds.items.ids())
    assert np.all(items.numbers() == np.arange(len(items)))
    assert items.vocabulary is ml_ds.items


@given(
    st.lists(integer_ids(), unique=True),
    st.lists(integer_ids(), unique=True),
)
def test_isin_integers(left, right):
    ill = ItemList(left)
    ilr = ItemList(right)

    mask = ill.isin(ilr)
    m2 = np.isin(np.array(left), np.array(right))
    assert np.all(mask == m2)


@given(
    st.lists(integer_ids(), unique=True),
    st.lists(integer_ids(), unique=True),
)
def test_isin_integer_vocab(left, right):
    vocab = Vocabulary(set(left + right))
    ill = ItemList(left, vocabulary=vocab)
    ilr = ItemList(right, vocabulary=vocab)

    mask = ill.isin(ilr)
    m2 = np.isin(np.array(left), np.array(right))
    assert np.all(mask == m2)


@given(st.lists(st.uuids(), unique=True), st.lists(st.uuids(), unique=True))
def test_isin_strings(left, right):
    left = [str(x) for x in left]
    right = [str(x) for x in right]

    ill = ItemList(left)
    ilr = ItemList(right)

    mask = ill.isin(ilr)
    m2 = np.isin(np.array(left), np.array(right))
    assert np.all(mask == m2)


@given(
    st.lists(st.uuids(), unique=True),
    st.lists(integer_ids(), unique=True),
)
def test_isin_strings_ints(left, right):
    left = [str(x) for x in left]

    ill = ItemList(left)
    ilr = ItemList(right)

    mask = ill.isin(ilr)
    assert not np.any(mask)


@given(
    st.lists(integer_ids(), unique=True),
    st.lists(st.uuids(), unique=True),
)
def test_isin_ints_strings(left, right):
    right = [str(x) for x in right]

    ill = ItemList(left)
    ilr = ItemList(right)

    mask = ill.isin(ilr)
    assert not np.any(mask)


@given(st.lists(integer_ids(), unique=True), st.lists(integer_ids(), unique=True))
def test_remove_ids(ids, remove):
    il = ItemList(ids, vocabulary=Vocabulary(ids))

    il2 = il.remove(ids=remove)
    print(repr(il2))

    intersect = set(ids) & set(remove)
    assert len(il2) == len(il) - len(intersect)
    assert not any(i in il2.ids() for i in remove)
    assert all(i in il2.ids() for i in ids if i not in remove)


@given(
    st.lists(integer_ids(), unique=True),
    st.lists(st.integers(min_value=0, max_value=20), unique=True),
)
def test_remove_numbers(ids, remove):
    il = ItemList(ids, vocabulary=Vocabulary(ids))

    il2 = il.remove(numbers=remove)

    intersect = set(il.numbers()) & set(remove)
    assert len(il2) == len(il) - len(intersect)
    assert not any(i in il2.numbers() for i in remove)
    assert all(i in il2.numbers() for i in il.numbers() if i not in remove)


@given(
    st.lists(st.integers(min_value=0, max_value=1000), unique=True),
    st.lists(st.integers(min_value=0, max_value=1000), unique=True),
)
def test_remove_numbers_no_vocab(ids, remove):
    il = ItemList(item_nums=ids)

    il2 = il.remove(numbers=remove)

    intersect = set(il.numbers()) & set(remove)
    assert len(il2) == len(il) - len(intersect)
    assert not any(i in il2.numbers() for i in remove)
    assert all(i in il2.numbers() for i in il.numbers() if i not in remove)


@given(
    st.lists(integer_ids(), unique=True),
    st.lists(st.integers(min_value=0, max_value=20), unique=True),
)
def test_remove_ids_no_vocab(ids, remove):
    il = ItemList(ids)

    il2 = il.remove(ids=remove)

    intersect = set(ids) & set(remove)
    assert len(il2) == len(il) - len(intersect)
    assert not any(i in il2.ids() for i in remove)
    assert all(i in il2.ids() for i in ids if i not in remove)


def test_repr_ids():
    il = ItemList(ITEMS, vocabulary=VOCAB)
    ilr = repr(il)
    print(ilr)
    assert re.search(r"ids: \['a'.*\]", ilr)
    assert re.search(r"numbers: <lazy>", ilr)


def test_repr_numbers():
    il = ItemList(item_nums=np.arange(len(ITEMS)), vocabulary=VOCAB)
    ilr = repr(il)
    print(ilr)
    assert re.search(r"numbers: \[0.*\]", ilr)
    assert re.search(r"ids: <lazy>", ilr)


@mark.skip("ranks are not stable")
def test_repr_lazy_rank():
    il = ItemList(item_nums=np.arange(len(ITEMS)), vocabulary=VOCAB, ordered=True)
    ilr = repr(il)
    print(ilr)
    assert re.search(r"numbers: \[0 .*\]", ilr)
    assert re.search(r"ids: <lazy>", ilr)
    assert re.search(r"rank: <lazy>", ilr)

    _rank = il.ranks()

    ilr = repr(il)
    print(ilr)
    assert re.search(r"numbers: \[0 .*\]", ilr)
    assert re.search(r"ids: <lazy>", ilr)
    assert re.search(r"rank: \[0 .*\]", ilr)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import subprocess as sp
from os import environ, fspath
from pathlib import Path

from pytest import mark, skip

from lenskit.data.dataset import Dataset
from lenskit.data.matrix import SparseRowArray
from lenskit.logging import get_logger
from lenskit.testing import ml_test_dir

# The LensKit versions we want to test backwards compatibility with
LK_VERSIONS = ["2025.1.1"]

_log = get_logger(__name__)
_ml_path = Path("data/ml-20m.zip")


@mark.slow
@mark.parametrize("version", LK_VERSIONS)
def test_data_backwards_compat(version, tmpdir: Path):
    "Test that we can load datasets prepared by old versions."
    _log.info("processing ML file", version=version)

    try:
        sp.call(["uvx", "--version"])
    except FileNotFoundError:
        skip("uvx not installed")

    out_path = tmpdir / "ml-small.lk"
    pkg = f"lenskit=={version}"

    sp.check_call(
        [
            "uvx",
            "--isolated",
            pkg,
            "data",
            "convert",
            "--movielens",
            fspath(ml_test_dir),
            fspath(out_path),
        ],
        env=environ | {"UV_TORCH_BACKEND": "cpu"},
    )

    _log.info("loading dataset")
    ml = Dataset.load(out_path)
    assert ml.schema.version is not None


@mark.realdata
@mark.skipif(not _ml_path.exists(), reason="ml-20m not available")
@mark.parametrize("version", LK_VERSIONS)
def test_data_backwards_ml20m(version, tmpdir: Path):
    "Test that we can load datasets prepared by old versions (ML20M)."
    try:
        sp.call(["uvx", "--version"])
    except FileNotFoundError:
        skip("uvx not installed")

    _log.info("processing ML file", version=version)

    out_path = tmpdir / "ml-20m"
    pkg = f"lenskit=={version}"

    sp.check_call(
        [
            "uvx",
            "--isolated",
            pkg,
            "data",
            "convert",
            "--movielens",
            fspath(_ml_path),
            fspath(out_path),
        ],
        env=environ | {"UV_TORCH_BACKEND": "cpu"},
    )

    _log.info("loading dataset")
    ml = Dataset.load(out_path)
    assert ml.schema.version is not None

    arr = ml.entities("item").attribute("tag_counts").arrow()
    assert isinstance(arr, SparseRowArray)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Basic batching support.
"""

from __future__ import annotations

import math
from collections.abc import Iterator
from dataclasses import dataclass

from lenskit.logging import get_logger

from .types import Extent

_log = get_logger(__name__)


@dataclass
class BatchIter:
    """
    Iterator over batches of a collection size.

    The length of this collection is the number of batches,
    and iterating over it yields batch extents.
    """

    n: int
    "The total number of items."
    size: int
    "The batch size."

    @property
    def batch_count(self) -> int:
        return math.ceil(self.n / self.size)

    def __len__(self) -> int:
        return self.batch_count

    def __iter__(self) -> Iterator[Extent]:
        _log.debug("iterating batches", N=self.n, count=self.batch_count, batch_size=self.size)
        for start in range(0, self.n, self.size):
            end = min(start + self.size, self.n)
            yield Extent(start, end)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import functools as ft
import itertools as it
import math

import numpy as np
import pandas as pd

import pytest

from lenskit.data import Dataset
from lenskit.splitting.records import crossfold_records, sample_records


def test_crossfold_records(ml_ds: Dataset):
    splits = crossfold_records(ml_ds, 5)
    splits = list(splits)
    assert len(splits) == 5

    for s in splits:
        # do we have all the data?
        test_count = s.test_size
        assert test_count + s.train.interaction_count == ml_ds.interactions().count()
        test_pairs = set((u, i) for (u, il) in s.test for i in il.ids())
        tdf = s.train.interaction_matrix(format="pandas", field="rating", original_ids=True)
        train_pairs = set(zip(tdf["user_id"], tdf["item_id"]))

        # no overlap
        assert not (test_pairs & train_pairs)
        # union is complete
        assert len(test_pairs | train_pairs) == ml_ds.interactions().count()

    # the test sets are pairwise disjoint
    for s1, s2 in it.product(splits, splits):
        if s1 is s2:
            continue

        p1 = set((u, i) for (u, il) in s1.test for i in il.ids())
        p2 = set((u, i) for (u, il) in s2.test for i in il.ids())
        assert not (p1 & p2)


def test_sample_records_once(ml_ds):
    split = sample_records(ml_ds, size=1000)

    test_count = split.test_size
    assert test_count == 1000
    assert test_count + split.train.interaction_count == ml_ds.interactions().count()
    test_pairs = set((u, i) for (u, il) in split.test for i in il.ids())
    tdf = split.train.interaction_matrix(format="pandas", field="rating", original_ids=True)
    train_pairs = set(zip(tdf["user_id"], tdf["item_id"]))

    # no overlap
    assert not (test_pairs & train_pairs)
    # union is complete
    assert len(test_pairs | train_pairs) == ml_ds.interactions().count()


def test_sample_records(ml_ds):
    splits = sample_records(ml_ds, size=1000, repeats=5)
    splits = list(splits)
    assert len(splits) == 5

    for s in splits:
        test_count = s.test_size
        assert test_count == 1000
        assert test_count + s.train.interaction_count == ml_ds.interactions().count()
        test_pairs = set((u, i) for (u, il) in s.test for i in il.ids())
        tdf = s.train.interaction_matrix(format="pandas", field="rating", original_ids=True)
        train_pairs = set(zip(tdf["user_id"], tdf["item_id"]))

        # no overlap
        assert not (test_pairs & train_pairs)
        # union is complete
        assert len(test_pairs | train_pairs) == ml_ds.interactions().count()

    for s1, s2 in it.product(splits, splits):
        if s1 is s2:
            continue

        p1 = set((u, i) for (u, il) in s1.test for i in il.ids())
        p2 = set((u, i) for (u, il) in s2.test for i in il.ids())
        assert not (p1 & p2)


def test_sample_records_test_only(ml_ds):
    splits = sample_records(ml_ds, size=1000, repeats=1, test_only=True)
    splits = list(splits)
    assert len(splits) == 1

    for s in splits:
        test_count = s.test_size
        assert test_count == 1000
        assert s.train.interaction_count == 0


def test_sample_rows_more_smaller_parts(ml_ds: Dataset):
    splits = sample_records(ml_ds, 500, repeats=10)
    splits = list(splits)
    assert len(splits) == 10

    for s in splits:
        test_count = s.test_size
        assert test_count == 500
        assert test_count + s.train.interaction_count == ml_ds.interactions().count()
        test_pairs = set((u, i) for (u, il) in s.test for i in il.ids())
        tdf = s.train.interaction_matrix(format="pandas", field="rating", original_ids=True)
        train_pairs = set(zip(tdf["user_id"], tdf["item_id"]))

        # no overlap
        assert not (test_pairs & train_pairs)
        # union is complete
        assert len(test_pairs | train_pairs) == ml_ds.interactions().count()

    for s1, s2 in it.product(splits, splits):
        if s1 is s2:
            continue

        p1 = set((u, i) for (u, il) in s1.test for i in il.ids())
        p2 = set((u, i) for (u, il) in s2.test for i in il.ids())
        assert not (p1 & p2)


def test_sample_non_disjoint(ml_ds: Dataset):
    splits = sample_records(ml_ds, 1000, repeats=10, disjoint=False)
    splits = list(splits)
    assert len(splits) == 10

    for s in splits:
        test_count = s.test_size
        assert test_count == 1000
        assert test_count + s.train.interaction_count == ml_ds.interactions().count()
        test_pairs = set((u, i) for (u, il) in s.test for i in il.ids())
        tdf = s.train.interaction_matrix(format="pandas", field="rating", original_ids=True)
        train_pairs = set(zip(tdf["user_id"], tdf["item_id"]))

        # no overlap
        assert not (test_pairs & train_pairs)
        # union is complete
        assert len(test_pairs | train_pairs) == ml_ds.interactions().count()

    # There are enough splits & items we should pick at least one duplicate
    ipairs = (
        (
            set((u, i) for (u, il) in s1.test for i in il.ids()),
            set((u, i) for (u, il) in s2.test for i in il.ids()),
        )
        for (s1, s2) in it.product(splits, splits)
    )
    isizes = [len(i1.intersection(i2)) for (i1, i2) in ipairs]
    assert any(n > 0 for n in isizes)


@pytest.mark.slow
def test_sample_oversize(ml_ds: Dataset):
    splits = sample_records(ml_ds, 10000, repeats=50)
    splits = list(splits)
    assert len(splits) == 50

    for s in splits:
        test_count = s.test_size
        assert test_count + s.train.interaction_count == ml_ds.interactions().count()
        test_pairs = set((u, i) for (u, il) in s.test for i in il.ids())
        tdf = s.train.interaction_matrix(format="pandas", field="rating", original_ids=True)
        train_pairs = set(zip(tdf["user_id"], tdf["item_id"]))

        # no overlap
        assert not (test_pairs & train_pairs)
        # union is complete
        assert len(test_pairs | train_pairs) == ml_ds.interactions().count()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
from typing import Generator, NamedTuple

import numpy as np
import pandas as pd

from pytest import approx, fixture, mark, warns

from lenskit.basic import BiasScorer, PopScorer
from lenskit.batch import BatchPipelineRunner, predict, recommend, score
from lenskit.data import Dataset, ItemList, UserIDKey, from_interactions_df
from lenskit.data.adapt import normalize_interactions_df
from lenskit.diagnostics import DataWarning
from lenskit.metrics import NDCG, RBP, RMSE, RunAnalysis
from lenskit.pipeline import Pipeline, topn_pipeline
from lenskit.splitting import SampleN, TTSplit, sample_users
from lenskit.testing import ml_100k, ml_ds, ml_ratings  # noqa: F401

_log = logging.getLogger(__name__)


class MLB(NamedTuple):
    ratings: pd.DataFrame
    data: Dataset
    pipeline: Pipeline


@fixture
def mlb(ml_ratings: pd.DataFrame, ml_ds: Dataset) -> MLB:
    bias = BiasScorer()
    pipeline = topn_pipeline(bias, predicts_ratings=True)
    pipeline.train(ml_ds)

    return MLB(
        normalize_interactions_df(ml_ratings),
        ml_ds,
        pipeline,
    )


@fixture
def ml_split(ml_100k: pd.DataFrame) -> Generator[TTSplit, None, None]:
    ds = from_interactions_df(ml_100k)
    yield sample_users(ds, 200, SampleN(5))


def test_predict_single(mlb: MLB):
    res = predict(mlb.pipeline, {1: ItemList([31])}, n_jobs=1)

    assert len(res) == 1
    uid, result = next(iter(res))
    assert isinstance(uid, UserIDKey)
    assert uid.user_id == 1
    assert len(result) == 1
    assert result.ids()[0] == 31

    preds = result.field("score")
    assert preds is not None
    assert preds >= 1 and preds <= 5


def test_score_single(mlb: MLB):
    res = score(mlb.pipeline, {1: ItemList([31])}, n_jobs=1)

    assert len(res) == 1
    uid, result = next(iter(res))
    assert isinstance(uid, UserIDKey)
    assert uid.user_id == 1
    assert len(result) == 1
    assert result.ids()[0] == 31

    preds = result.field("score")
    assert preds is not None
    assert preds >= 1 and preds <= 5


def test_recommend_user(mlb: MLB):
    user = 5

    results = recommend(mlb.pipeline, [user], n=10, n_jobs=1)

    assert len(results) == 1

    uid, ranking = next(iter(results))
    assert isinstance(uid, UserIDKey)
    assert uid.user_id == user
    assert isinstance(ranking, ItemList)
    assert ranking.ordered
    assert len(ranking) == 10

    # they should be in decreasing order
    score = ranking.scores()
    assert score is not None
    assert all(np.diff(score) <= 0)


@mark.parametrize(("ncpus"), [None, 1, 2])
@mark.eval
def test_bias_batch(ml_split: TTSplit, ncpus: int | None):
    algo = BiasScorer(damping=5)
    pipeline = topn_pipeline(algo, predicts_ratings=True, n=20)
    pipeline.train(ml_split.train)

    runner = BatchPipelineRunner(n_jobs=ncpus)
    runner.recommend()
    runner.predict()

    results = runner.run(pipeline, ml_split.test)

    preds = results.output("predictions")

    pa = RunAnalysis()
    pa.add_metric(RMSE())
    pred_acc = pa.measure(preds, ml_split.test)
    pas = pred_acc.list_summary()
    print(pas)
    assert pas.loc["RMSE", "mean"] == approx(0.949, rel=0.1)

    recs = results.output("recommendations")
    ra = RunAnalysis()
    ra.add_metric(NDCG())
    ra.add_metric(RBP())
    rec_acc = ra.measure(recs, ml_split.test)
    ras = rec_acc.list_summary()
    print(ras)

    assert ras.loc["RBP", "mean"] > 0
    assert ras.loc["NDCG", "mean"] > 0


@mark.parametrize("ncpus", [None, 1, 2])
@mark.eval
def test_pop_batch_recommend(ml_split: TTSplit, ncpus: int | None):
    algo = PopScorer()
    pipeline = topn_pipeline(algo, n=20)
    pipeline.train(ml_split.train)

    runner = BatchPipelineRunner(n_jobs=ncpus)
    runner.recommend()

    results = runner.run(pipeline, ml_split.test)

    recs = results.output("recommendations")
    ra = RunAnalysis()
    ra.add_metric(NDCG())
    ra.add_metric(RBP())
    rec_acc = ra.measure(recs, ml_split.test)
    ras = rec_acc.list_summary()
    print(ras)

    assert ras.loc["RBP", "mean"] > 0
    assert ras.loc["NDCG", "mean"] > 0


@mark.eval
def test_bias_df(ml_split: TTSplit):
    algo = BiasScorer()
    pipeline = topn_pipeline(algo, predicts_ratings=True, n=20)
    pipeline.train(ml_split.train)

    runner = BatchPipelineRunner()
    runner.recommend()

    with warns(DataWarning):
        results = runner.run(pipeline, ml_split.test.to_df())

    recs = results.output("recommendations")
    ra = RunAnalysis()
    ra.add_metric(NDCG())
    ra.add_metric(RBP())
    rec_acc = ra.measure(recs, ml_split.test)
    ras = rec_acc.list_summary()
    print(ras)

    assert ras.loc["RBP", "mean"] > 0
    assert ras.loc["NDCG", "mean"] > 0



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Mapping
from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class ParameterContainer(Protocol):  # pragma: nocover
    """
    Protocol for components with learned parameters to enable saving, reloading,
    checkpointing, etc.

    Components that learn parameters from training data should implement this
    protocol, and also work when pickled or pickled.  Pickling is sometimes used
    for convenience, but parameter / state dictionaries allow serializing wtih
    tools like ``safetensors`` or ``zarr``.

    Initializing a component with the same configuration as a trained component,
    and loading its parameters with :meth:`load_parameters`, should result in a
    component that is functionally equivalent to the original trained component.

    Stability:
        Experimental
    """

    def get_parameters(self) -> Mapping[str, object]:
        """
        Get the component's parameters.

        Returns:
            The model's parameters, as a dictionary from names to parameter data
            (usually arrays, tensors, etc.).
        """
        raise NotImplementedError()

    def load_parameters(self, state: Mapping[str, Any]) -> None:
        """
        Reload model state from parameters saved via :meth:`get_parameters`.

        Args:
            params:
                The model parameters, as a dictionary from names to parameter
                data (arrays, tensors, etc.), as returned from
                :meth:`get_parameters`.
        """
        raise NotImplementedError()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

from lenskit.tuning import PipelineTuner
from lenskit.tuning.spec import TuningSpec


def test_tuner_spec():
    spec = TuningSpec.load(Path("pipelines/iknn-explicit-search.toml"))
    assert len(spec.space) == 1
    assert spec.component_name == "scorer"


def test_tuner_space():
    spec = TuningSpec.load(Path("pipelines/iknn-explicit-search.toml"))
    tuner = PipelineTuner(spec)

    space = tuner.search_space()
    assert space is not None



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import functools as ft
import itertools as it
import math

import numpy as np
import pandas as pd

import pytest

from lenskit.data import Dataset, from_interactions_df
from lenskit.splitting.holdout import SampleFrac, SampleN
from lenskit.splitting.users import crossfold_users, sample_users


def test_crossfold_users(ml_ds: Dataset):
    splits = crossfold_users(ml_ds, 5, SampleN(5))
    splits = list(splits)
    assert len(splits) == 5

    users = set()
    for s in splits:
        assert all(len(il) for il in s.test.lists())
        assert not any(u in users for u in s.test.keys())
        users |= set(k.user_id for k in s.test.keys())

        test_pairs = set((u, i) for (u, il) in s.test for i in il.ids())
        tdf = s.train.interaction_matrix(format="pandas", field="rating", original_ids=True)
        train_pairs = set(zip(tdf["user_id"], tdf["item_id"]))
        assert not test_pairs & train_pairs
        assert s.test_size + s.train.interactions().count() == ml_ds.interactions().count()

    assert users == set(ml_ds.users)


def test_crossfold_may_skip_train(ml_ratings: pd.DataFrame):
    "Partitioning when users may not have enough ratings to be in the train and test sets."
    # make a data set where some users only have 1 rating
    ml_ratings = ml_ratings.sample(frac=0.1)
    ucounts = ml_ratings.groupby("user_id")["rating"].count()
    assert ucounts.min() == 1  # we should have some small users!
    ucounts.name = "ur_count"
    ml_ds = from_interactions_df(ml_ratings)

    splits = crossfold_users(ml_ds, 5, SampleN(1))
    splits = list(splits)
    assert len(splits) == 5

    # now we go make sure we're missing some users! And don't have any NaN ml_ratings
    for split in splits:
        for u in ucounts[ucounts == 1].index:
            if u in split.test:
                row = split.train.user_row(u)
                assert row is not None
                assert len(row) == 0


def test_crossfold_users_frac(ml_ds: Dataset):
    splits = crossfold_users(ml_ds, 5, SampleFrac(0.2))
    splits = list(splits)
    assert len(splits) == 5
    ustats = ml_ds.user_stats()
    uss = ustats["count"] * 0.2

    for s in splits:
        assert all(len(il) >= uss.loc[u.user_id] - 1 for (u, il) in s.test)
        assert all(len(il) <= uss.loc[u.user_id] + 1 for (u, il) in s.test)
        assert s.test_size + s.train.interactions().count() == ml_ds.interactions().count()


def test_sample_users_single(ml_ds: Dataset):
    split = sample_users(ml_ds, 100, SampleN(5))

    assert len(split.test) == 100
    assert split.test_size == 500

    test_pairs = set((u, i) for (u, il) in split.test for i in il.ids())
    assert len(test_pairs) == split.test_size
    tdf = split.train.interaction_matrix(format="pandas", field="rating", original_ids=True)
    train_pairs = set(zip(tdf["user_id"], tdf["item_id"]))
    assert len(train_pairs) == split.train.interactions().count()
    assert len(test_pairs & train_pairs) == 0
    assert split.test_size + split.train.interactions().count() == ml_ds.interactions().count()


def test_sample_users(ml_ds: Dataset):
    splits = sample_users(ml_ds, 100, SampleN(5), repeats=5)
    splits = list(splits)
    assert len(splits) == 5

    aus = set()
    for s in splits:
        assert len(s.test) == 100
        assert s.test_size == 500
        # users are disjoint
        assert not any(u in aus for u in s.test.keys())
        aus |= set(k.user_id for k in s.test.keys())

        test_pairs = set((u, i) for (u, il) in s.test for i in il.ids())
        assert len(test_pairs) == s.test_size
        tdf = s.train.interaction_matrix(format="pandas", field="rating", original_ids=True)
        train_pairs = set(zip(tdf["user_id"], tdf["item_id"]))
        assert len(train_pairs) == s.train.interactions().count()
        assert len(test_pairs & train_pairs) == 0
        assert s.test_size + s.train.interactions().count() == ml_ds.interactions().count()


def test_sample_test_only(ml_ds: Dataset):
    splits = sample_users(ml_ds, 100, SampleN(5), repeats=1, test_only=True)
    splits = list(splits)
    assert len(splits) == 1

    for s in splits:
        assert len(s.test) == 100
        assert s.test_size == 500
        assert s.train.interaction_count == 0


def test_sample_users_non_disjoint(ml_ds: Dataset):
    splits = sample_users(ml_ds, 100, SampleN(5), repeats=5, disjoint=False)
    splits = list(splits)
    assert len(splits) == 5

    aus = set()

    for s in splits:
        assert len(s.test) == 100
        assert s.test_size == 500
        aus |= set(k.user_id for k in s.test.keys())

        test_pairs = set((u, i) for (u, il) in s.test for i in il.ids())
        assert len(test_pairs) == s.test_size
        tdf = s.train.interaction_matrix(format="pandas", field="rating", original_ids=True)
        train_pairs = set(zip(tdf["user_id"], tdf["item_id"]))
        assert len(train_pairs) == s.train.interactions().count()
        assert len(test_pairs & train_pairs) == 0
        assert s.test_size + s.train.interactions().count() == ml_ds.interactions().count()

    # some user appears at least once
    assert len(aus) < 500


@pytest.mark.slow
def test_sample_users_frac_oversize(ml_ds: Dataset):
    splits = sample_users(ml_ds, 100, SampleN(5), repeats=20)
    splits = list(splits)
    assert len(splits) == 20

    for s in splits:
        assert len(s.test) < 100
        assert all(len(il) == 5 for il in s.test.lists())


def test_sample_users_frac_oversize_ndj(ml_ds: Dataset):
    splits = sample_users(ml_ds, 100, SampleN(5), repeats=20, disjoint=False)
    splits = list(splits)
    assert len(splits) == 20

    for s in splits:
        assert len(s.test) == 100
        assert s.test_size == 5 * 100
        assert all([len(il) for il in s.test.lists()])



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import datetime as dt

import numpy as np

from pytest import approx

from lenskit.data import Dataset
from lenskit.splitting import split_global_time, split_temporal_fraction


def test_temporal_split_filtered_users(ml_ds: Dataset):
    point = dt.datetime.fromisoformat("2015-01-01")

    split = split_global_time(ml_ds, "2015-01-01", filter_test_users=True)
    assert np.all(split.train.interaction_table(format="pandas")["timestamp"] < point)

    train_users = split.train.user_stats().index[split.train.user_stats()["count"] > 0]

    for u, il in split.test:
        ts = il.field("timestamp", format="pandas")

        assert ts is not None
        assert np.all(ts >= point)
        assert u.user_id in train_users


def test_temporal_split_filtered_users_fraction(ml_ds: Dataset):
    split = split_temporal_fraction(ml_ds, 0.2, filter_test_users=True)

    point = min(il.field("timestamp").min() for il in split.test.lists())
    assert np.all(split.train.interaction_table(format="pandas")["timestamp"] < point)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle

import numpy as np
import pandas as pd

from pytest import approx, mark, skip

from lenskit.als import BiasedMFScorer
from lenskit.data import Dataset, ItemList, RecQuery, from_interactions_df, load_movielens_df
from lenskit.metrics import quick_measure_model
from lenskit.testing import BasicComponentTests, ScorerTests

_log = logging.getLogger(__name__)

simple_df = pd.DataFrame(
    {"item": [1, 1, 2, 3], "user": [10, 12, 10, 13], "rating": [4.0, 3.0, 5.0, 2.0]}
)
simple_ds = from_interactions_df(simple_df)


class TestExplicitALS(BasicComponentTests, ScorerTests):
    component = BiasedMFScorer
    expected_rmse = (0.89, 0.99)

    def verify_models_equivalent(self, orig, copy):
        assert copy.bias.global_bias == orig.bias.global_bias
        assert np.all(copy.bias.user_biases == orig.bias.user_biases)
        assert np.all(copy.bias.item_biases == orig.bias.item_biases)
        assert np.all(copy.user_embeddings == orig.user_embeddings)
        assert np.all(copy.item_embeddings == orig.item_embeddings)
        assert np.all(copy.items.index == orig.items.index)
        assert np.all(copy.users.index == orig.users.index)


def test_als_basic_build():
    algo = BiasedMFScorer(features=20, epochs=10)
    algo.train(simple_ds)

    assert algo.bias is not None
    assert algo.users is not None
    assert algo.user_embeddings is not None

    assert algo.bias.global_bias == approx(simple_df.rating.mean())
    assert set(algo.users.ids()) == set([10, 12, 13])
    assert set(algo.items.ids()) == set([1, 2, 3])
    assert algo.user_embeddings.shape == (3, 20)
    assert algo.item_embeddings.shape == (3, 20)

    assert algo.config.embedding_size == 20
    assert len(algo.users) == 3
    assert len(algo.items) == 3


def test_als_predict_basic():
    algo = BiasedMFScorer(features=20, epochs=10)
    algo.train(simple_ds)

    assert algo.bias is not None

    assert algo.bias.global_bias == approx(simple_df.rating.mean())

    preds = algo(query=10, items=ItemList([3]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert preds.loc[3] >= -0.1
    assert preds.loc[3] <= 5.1


def test_als_predict_basic_for_new_ratings():
    algo = BiasedMFScorer(features=20, epochs=10)
    algo.train(simple_ds)

    assert algo.bias is not None
    assert algo.bias.global_bias == approx(simple_df.rating.mean())

    query = RecQuery(15, ItemList(item_ids=[1, 2], rating=[4.0, 5.0]))
    preds = algo(query, items=ItemList([3]))

    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert preds.loc[3] >= -0.1
    assert preds.loc[3] <= 5.1


def test_als_predict_basic_for_new_user_with_new_ratings():
    u = 10
    i = 3

    algo = BiasedMFScorer(features=20, epochs=10)
    algo.train(simple_ds)

    preds = algo(query=u, items=ItemList([i]))
    preds = preds.scores("pandas", index="ids")
    assert preds is not None

    query = RecQuery(-1, ItemList(item_ids=[1, 2], rating=[4.0, 5.0]))

    new_preds = algo(query=query, items=ItemList([i]))
    new_preds = new_preds.scores("pandas", index="ids")
    assert new_preds is not None

    assert preds.loc[i] == approx(new_preds.loc[i], rel=9e-2)


def test_als_predict_for_new_users_with_new_ratings(rng, ml_ds: Dataset):
    n_users = 3
    n_items = 2

    users = rng.choice(ml_ds.users.ids(), n_users)
    items = rng.choice(ml_ds.items.ids(), n_items)

    algo = BiasedMFScorer(features=20, epochs=10)
    algo.train(ml_ds)

    _log.debug("Items: " + str(items))
    assert algo.bias is not None
    assert algo.users is not None
    assert algo.user_embeddings is not None

    for u in users:
        _log.debug(f"user: {u}")
        preds = algo(query=u, items=ItemList(items))
        assert np.all(preds.ids() == items)

        user_data = ml_ds.user_row(u)

        _log.debug("user_features from fit: " + str(algo.user_embeddings[algo.users.number(u), :]))

        query = RecQuery(-1, user_data)
        new_preds = algo(query=query, items=ItemList(items))
        assert np.all(new_preds.ids() == items)

        _log.debug("preds: " + str(preds.scores()))
        _log.debug("new_preds: " + str(new_preds.scores()))
        _log.debug("------------")
        assert new_preds.scores() == approx(preds.scores(), rel=9e-2)


def test_als_predict_bad_item():
    algo = BiasedMFScorer(features=20, epochs=10)
    algo.train(simple_ds)

    assert algo.bias is not None
    assert algo.bias.global_bias == approx(simple_df.rating.mean())

    preds = algo(query=10, items=ItemList([4]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 4
    assert np.isnan(preds.loc[4])


def test_als_predict_bad_user():
    algo = BiasedMFScorer(features=20, epochs=10)
    algo.train(simple_ds)

    assert algo.bias is not None
    assert algo.bias.global_bias == approx(simple_df.rating.mean())

    preds = algo(query=50, items=ItemList([3]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert np.isnan(preds.loc[3])


def test_als_predict_no_user_features_basic(rng: np.random.Generator, ml_ds: Dataset):
    n_items = 2

    u = rng.choice(ml_ds.users.ids(), 1).item()
    items = rng.choice(ml_ds.items.ids(), n_items)

    algo = BiasedMFScorer(features=5, epochs=10)
    algo.train(ml_ds)
    _log.debug("Items: " + str(items))
    assert algo.bias is not None
    assert algo.users is not None
    assert algo.user_embeddings is not None

    algo_no_user_features = BiasedMFScorer(features=5, epochs=10, user_embeddings=False)
    algo_no_user_features.train(ml_ds)

    assert algo_no_user_features.user_embeddings is None

    _log.debug(f"user: {u}")
    preds = algo(query=u, items=ItemList(item_ids=items))

    user_data = ml_ds.user_row(u)

    _log.debug("user_features from fit: " + str(algo.user_embeddings[algo.users.number(u), :]))

    query = RecQuery(-1, user_data)
    new_preds = algo_no_user_features(query, items=ItemList(items))

    _log.debug("preds: " + str(preds.scores()))
    _log.debug("new_preds: " + str(new_preds.scores()))
    _log.debug("------------")
    assert new_preds.scores() == approx(preds.scores(), rel=9e-1)


@mark.slow
def test_als_train_large(ml_ratings, ml_ds: Dataset):
    algo = BiasedMFScorer(features=20, epochs=10)
    algo.train(ml_ds)

    assert algo.bias is not None
    assert algo.users is not None
    assert algo.user_embeddings is not None

    assert algo.bias.global_bias == approx(ml_ratings.rating.mean())
    assert algo.config.embedding_size == 20
    assert len(algo.items) == ml_ds.item_count
    assert len(algo.users) == ml_ds.user_count

    ratings = ml_ds.interaction_matrix(format="pandas")
    gmean = ratings["rating"].mean()

    istats = ml_ds.item_stats()
    icounts = istats["rating_count"]
    isums = istats["mean_rating"] * icounts
    is2 = isums - icounts * gmean
    imeans = is2 / (icounts + 5)
    ibias = pd.Series(algo.bias.item_biases, index=algo.items.index)
    imeans, ibias = imeans.align(ibias, fill_value=0.0)
    assert ibias.values == approx(imeans.values, rel=1.0e-3)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging

import numpy as np
import pandas as pd

from pytest import approx

from lenskit.data import ItemList
from lenskit.metrics.ranking import Recall
from lenskit.testing import demo_recs  # noqa: F401

_log = logging.getLogger(__name__)


def _test_recall(items, rel, **kwargs):
    recs = ItemList(items, ordered=True)
    truth = ItemList(rel)
    return Recall(**kwargs).measure_list(recs, truth)


def test_recall_empty_zero():
    prec = _test_recall([], [1, 3])
    assert prec == approx(0)


def test_recall_norel_na():
    prec = _test_recall([1, 3], [])
    assert np.isnan(prec)


def test_recall_simple_cases():
    prec = _test_recall([1, 3], [1, 3])
    assert prec == approx(1.0)

    prec = _test_recall([1], [1, 3])
    assert prec == approx(0.5)

    prec = _test_recall([1, 2, 3, 4], [1, 3])
    assert prec == approx(1.0)

    prec = _test_recall([1, 2, 3, 4], [1, 3, 5])
    assert prec == approx(2.0 / 3)

    prec = _test_recall([1, 2, 3, 4], range(5, 10))
    assert prec == approx(0.0)

    prec = _test_recall([1, 2, 3, 4], range(4, 9))
    assert prec == approx(0.2)


def test_recall_series():
    prec = _test_recall(pd.Series([1, 3]), pd.Series([1, 3]))
    assert prec == approx(1.0)

    prec = _test_recall(pd.Series([1, 2, 3]), pd.Series([1, 3, 5, 7]))
    assert prec == approx(0.5)

    prec = _test_recall(pd.Series([1, 2, 3, 4]), pd.Series(range(4, 9)))
    assert prec == approx(0.2)


def test_recall_series_set():
    prec = _test_recall(pd.Series([1, 2, 3, 4]), [1, 3, 5, 7])
    assert prec == approx(0.5)

    prec = _test_recall(pd.Series([1, 2, 3, 4]), range(4, 9))
    assert prec == approx(0.2)


def test_recall_series_index():
    prec = _test_recall(pd.Series([1, 3]), pd.Index([1, 3]))
    assert prec == approx(1.0)

    prec = _test_recall(pd.Series([1, 2, 3, 4]), pd.Index([1, 3, 5, 7]))
    assert prec == approx(0.5)

    prec = _test_recall(pd.Series([1, 2, 3, 4]), pd.Index(range(4, 9)))
    assert prec == approx(0.2)


def test_recall_series_array():
    prec = _test_recall(pd.Series([1, 3]), np.array([1, 3]))
    assert prec == approx(1.0)

    prec = _test_recall(pd.Series([1, 2, 3, 4]), np.array([1, 3, 5, 7]))
    assert prec == approx(0.5)

    prec = _test_recall(pd.Series([1, 2, 3, 4]), np.arange(4, 9, 1, "u4"))
    assert prec == approx(0.2)


def test_recall_array():
    prec = _test_recall(np.array([1, 3]), np.array([1, 3]))
    assert prec == approx(1.0)

    prec = _test_recall(np.array([1, 2, 3, 4]), np.array([1, 3, 5, 7]))
    assert prec == approx(0.5)

    prec = _test_recall(np.array([1, 2, 3, 4]), np.arange(4, 9, 1, "u4"))
    assert prec == approx(0.2)


def test_recall_long_rel():
    rel = np.arange(100)
    items = [1, 0, 150, 3, 10]

    r = _test_recall(items, rel, k=5)
    assert r == approx(0.8)


def test_recall_long_items():
    rel = np.arange(100)
    items = [1, 0, 150, 3, 10, 30, 120, 4, 17]

    r = _test_recall(items, rel, k=5)
    assert r == approx(0.8)


def test_recall_partial_rel():
    rel = np.arange(100)
    items = [1, 0, 150, 3, 10]

    r = _test_recall(items, rel, k=10)
    assert r == approx(0.4)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from dataclasses import dataclass

from pydantic import BaseModel
from pydantic.dataclasses import dataclass as pydantic_dataclass

from pytest import mark

from lenskit.pipeline import Component, PipelineBuilder
from lenskit.pipeline.cache import PipelineCache
from lenskit.pipeline.nodes import ComponentInstanceNode


@dataclass
class PrefixConfigDC:
    prefix: str = "UNDEFINED"


class PrefixConfigM(BaseModel):
    prefix: str = "UNDEFINED"


@pydantic_dataclass
class PrefixConfigPYDC:
    prefix: str = "UNDEFINED"


class PrefixerDC(Component[str]):
    config: PrefixConfigDC

    def __call__(self, msg: str) -> str:
        return self.config.prefix + msg


class PrefixerM(Component[str]):
    config: PrefixConfigM

    def __call__(self, msg: str) -> str:
        return self.config.prefix + msg


class PrefixerPYDC(Component[str]):
    config: PrefixConfigPYDC

    def __call__(self, msg: str) -> str:
        return self.config.prefix + msg


@dataclass
class SuffixConfig:
    suffix: str


class Suffixer(Component[str]):
    config: SuffixConfig

    def __call__(self, msg: str) -> str:
        return msg + self.config.suffix


def lowercase(msg: str) -> str:
    "Lowercase a string."
    return msg.lower()


@mark.parametrize("prefixer", [PrefixerDC, PrefixerM, PrefixerPYDC])
def test_config_reuse_only(prefixer: type[Component]):
    ccls = prefixer.config_class()  # type: ignore
    assert ccls is not None

    build = PipelineBuilder("test")
    msg = build.create_input("msg", str)
    build.add_component("prefix", prefixer, ccls(prefix="hello "), msg=msg)

    cache = PipelineCache()
    p1 = build.build(cache)
    p2 = build.build(cache)

    n1 = p1.node("prefix")
    assert isinstance(n1, ComponentInstanceNode)
    c1 = n1.component

    n2 = p2.node("prefix")
    assert isinstance(n2, ComponentInstanceNode)
    c2 = n2.component

    assert c1 is c2

    assert p1.run("prefix", msg="woozle") == "hello woozle"


@mark.parametrize("prefixer", [PrefixerDC, PrefixerM, PrefixerPYDC])
def test_config_reuse_one(prefixer: type[Component]):
    ccls = prefixer.config_class()  # type: ignore
    assert ccls is not None

    build = PipelineBuilder("test")
    msg = build.create_input("msg", str)
    pm = build.add_component("prefix", prefixer, ccls(prefix="hello "), msg=msg)
    build.add_component("suffix", Suffixer, {"suffix": " for now"}, msg=pm)

    cache = PipelineCache()

    p1 = build.build(cache)

    build.replace_component("prefix", prefixer, ccls(prefix="goodbye "))
    p2 = build.build(cache)

    n1 = p1.node("prefix")
    assert isinstance(n1, ComponentInstanceNode)
    c1 = n1.component

    n2 = p2.node("prefix")
    assert isinstance(n2, ComponentInstanceNode)
    c2 = n2.component

    assert c1 is not c2

    n1 = p1.node("suffix")
    assert isinstance(n1, ComponentInstanceNode)
    c1 = n1.component

    n2 = p2.node("suffix")
    assert isinstance(n2, ComponentInstanceNode)
    c2 = n2.component

    assert c1 is c2

    assert p1.run("suffix", msg="woozle") == "hello woozle for now"
    assert p2.run("suffix", msg="woozle") == "goodbye woozle for now"



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
from typing_extensions import override

from lenskit.data.items import ItemList

from ._base import ListMetric, RankingMetricBase


class AveragePrecision(ListMetric, RankingMetricBase):
    """
    Compute Average Precision (AP) for a single user's recommendations.  This is
    the average of the precision at each relevant item in the ranked list.
    """

    @property
    def label(self):
        if self.k is not None:
            return f"AveragePrecision@{self.k}"
        else:
            return "AveragePrecision"

    @override
    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        recs = self.truncate(recs)
        nrecs = len(recs)
        if nrecs == 0:
            return np.nan

        ap_sum = 0.0
        good = recs.isin(test)
        sum_good = np.cumsum(good)  # cumulative count of “good” up to each rank
        ranks = recs.ranks()
        assert ranks is not None
        ap_sum = np.sum(sum_good[good] / ranks[good]).item()

        denom = min(len(test), len(recs))
        return ap_sum / denom



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any

from pydantic import BaseModel, TypeAdapter
from pydantic.dataclasses import dataclass as pydantic_dataclass

from pytest import mark

from lenskit.pipeline import PipelineBuilder
from lenskit.pipeline.components import Component, ComponentConstructor
from lenskit.pipeline.nodes import ComponentConstructorNode


class EarlyConfig(Component[str]):
    config: PrefixConfigDC

    def __call__(self, msg: str) -> str:
        return self.config.prefix + msg


@dataclass
class PrefixConfigDC:
    prefix: str = "UNDEFINED"


class PrefixConfigM(BaseModel):
    prefix: str = "UNDEFINED"


@pydantic_dataclass
class PrefixConfigPYDC:
    prefix: str = "UNDEFINED"


class PrefixerDC(Component[str]):
    config: PrefixConfigDC

    def __call__(self, msg: str) -> str:
        return self.config.prefix + msg


class PrefixerM(Component[str]):
    config: PrefixConfigM

    def __call__(self, msg: str) -> str:
        return self.config.prefix + msg


# make sure it works with sub-sub-classes
class PrefixerM2(PrefixerM):
    config: PrefixConfigM


class PrefixerPYDC(Component[str]):
    config: PrefixConfigPYDC

    def __call__(self, msg: str) -> str:
        return self.config.prefix + msg


@mark.parametrize("prefixer", [PrefixerDC, PrefixerM, PrefixerPYDC, PrefixerM2])
def test_config_setup(prefixer: type[Component]):
    ccls = prefixer.config_class()  # type: ignore
    assert ccls is not None

    comp = prefixer()
    assert isinstance(comp.config, ccls)


@mark.parametrize("prefixer", [PrefixerDC, PrefixerM, PrefixerPYDC])
def test_auto_config_roundtrip(prefixer: type[Component]):
    comp = prefixer(prefix="FOOBIE BLETCH")

    cfg = comp.config
    cfg_data = comp.dump_config()
    assert "prefix" in cfg_data

    c2 = prefixer(cfg)
    assert c2 is not comp
    assert c2.config.prefix == comp.config.prefix

    c3 = prefixer(prefixer.validate_config(cfg_data))
    assert c3 is not comp
    assert c3.config.prefix == comp.config.prefix


@mark.parametrize("prefixer", [PrefixerDC, PrefixerM, PrefixerPYDC])
def test_pipeline_config(prefixer: ComponentConstructor[Any, str]):
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)
    pn = pipe.add_component("prefix", prefixer, {"prefix": "scroll named "}, msg=msg)
    assert isinstance(pn, ComponentConstructorNode)
    assert pn.constructor == prefixer
    assert getattr(pn.config, "prefix") == "scroll named "

    pipe = pipe.build()
    assert pipe.run("prefix", msg="FOOBIE BLETCH") == "scroll named FOOBIE BLETCH"

    config = pipe.config.components
    print(TypeAdapter(dict).dump_json(config, indent=2))

    assert "prefix" in config
    assert config["prefix"].config
    assert config["prefix"].config["prefix"] == "scroll named "


@mark.parametrize("prefixer", [PrefixerDC, PrefixerM, PrefixerPYDC])
def test_pipeline_config_roundtrip(prefixer: type[Component]):
    comp = prefixer(prefix="scroll named ")

    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)
    pipe.add_component("prefix", comp, msg=msg)
    pipe.default_component("prefix")

    assert pipe.build().run("prefix", msg="FOOBIE BLETCH") == "scroll named FOOBIE BLETCH"

    config = pipe.build_config()
    print(config.model_dump_json(indent=2))

    p2 = PipelineBuilder.from_config(config)
    assert p2.node("prefix", missing="none") is not None
    assert p2.build().run(msg="READ ME") == "scroll named READ ME"



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import functools as ft
import itertools as it
import math

import numpy as np
import pandas as pd

import pytest

from lenskit.data import Dataset
from lenskit.splitting.holdout import LastFrac, LastN, SampleFrac, SampleN


def test_sample_n(ml_ds: Dataset):
    users = np.random.choice(ml_ds.users.ids(), 5, replace=False)

    s5 = SampleN(5)
    for u in users:
        row = ml_ds.user_row(u)
        assert row is not None
        tst = s5(row)
        mask = np.isin(row.ids(), tst.ids())
        trn = row[~mask]
        assert len(tst) == 5
        assert len(tst) + len(trn) == len(row)

    s10 = SampleN(10)
    for u in users:
        row = ml_ds.user_row(u)
        assert row is not None
        tst = s10(row)
        mask = np.isin(row.ids(), tst.ids())
        trn = row[~mask]
        assert len(tst) == 10
        assert len(tst) + len(trn) == len(row)


def test_sample_frac(ml_ds: Dataset):
    users = np.random.choice(ml_ds.users.ids(), 5, replace=False)

    samp = SampleFrac(0.2)
    for u in users:
        row = ml_ds.user_row(u)
        assert row is not None
        tst = samp(row)
        mask = np.isin(row.ids(), tst.ids())
        trn = row[~mask]
        assert len(tst) + len(trn) == len(row)
        assert len(tst) >= math.floor(len(row) * 0.2)
        assert len(tst) <= math.ceil(len(row) * 0.2)

    samp = SampleFrac(0.5)
    for u in users:
        row = ml_ds.user_row(u)
        assert row is not None
        tst = samp(row)
        mask = np.isin(row.ids(), tst.ids())
        trn = row[~mask]
        assert len(tst) + len(trn) == len(row)
        assert len(tst) >= math.floor(len(row) * 0.5)
        assert len(tst) <= math.ceil(len(row) * 0.5)


def test_last_n(ml_ds: Dataset):
    users = np.random.choice(ml_ds.users.ids(), 5, replace=False)

    samp = LastN(5)
    for u in users:
        row = ml_ds.user_row(u)
        assert row is not None
        tst = samp(row)
        mask = np.isin(row.ids(), tst.ids())
        trn = row[~mask]
        assert len(tst) == 5
        assert len(tst) + len(trn) == len(row)
        assert tst.field("timestamp").min() >= trn.field("timestamp").max()  # type: ignore

    samp = LastN(7)
    for u in users:
        row = ml_ds.user_row(u)
        assert row is not None
        tst = samp(row)
        mask = np.isin(row.ids(), tst.ids())
        trn = row[~mask]
        assert len(tst) == 7
        assert len(tst) + len(trn) == len(row)
        assert tst.field("timestamp").min() >= trn.field("timestamp").max()  # type: ignore


def test_last_frac(ml_ds: Dataset):
    users = np.random.choice(ml_ds.users.ids(), 5, replace=False)

    samp = LastFrac(0.2, "timestamp")
    for u in users:
        row = ml_ds.user_row(u)
        assert row is not None
        tst = samp(row)
        mask = np.isin(row.ids(), tst.ids())
        trn = row[~mask]
        assert len(tst) + len(trn) == len(row)
        assert len(tst) >= math.floor(len(row) * 0.2)
        assert len(tst) <= math.ceil(len(row) * 0.2)
        assert tst.field("timestamp").min() >= trn.field("timestamp").max()  # type: ignore

    samp = LastFrac(0.5, "timestamp")
    for u in users:
        row = ml_ds.user_row(u)
        assert row is not None
        tst = samp(row)
        mask = np.isin(row.ids(), tst.ids())
        trn = row[~mask]
        assert len(tst) + len(trn) == len(row)
        assert len(tst) >= math.floor(len(row) * 0.5)
        assert len(tst) <= math.ceil(len(row) * 0.5)
        assert tst.field("timestamp").min() >= trn.field("timestamp").max()  # type: ignore



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle

import numpy as np
import pandas as pd
from numpy.typing import NDArray

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import given, settings

from lenskit.basic import PopScorer
from lenskit.basic.topn import TopNRanker
from lenskit.data.items import ItemList

_log = logging.getLogger(__name__)


@st.composite
def scored_lists(draw: st.DrawFn) -> ItemList:
    n = draw(st.integers(0, 1000))
    ids = np.arange(1, n + 1, dtype=np.int32)
    scores = draw(nph.arrays(nph.floating_dtypes(endianness="=", sizes=[32, 64]), n))
    return ItemList(item_ids=ids, scores=scores)


@given(scored_lists())
def test_unlimited_ranking(items: ItemList):
    topn = TopNRanker()
    ranked = topn(items=items)

    ids = items.ids()
    scores = items.scores("numpy")
    assert scores is not None
    invalid = np.isnan(scores)
    _log.info("ranking %d items, %d invalid", len(ids), np.sum(invalid))

    assert isinstance(ranked, ItemList)
    assert len(ranked) <= len(items)
    assert ranked.ordered
    # all valid items are included
    assert len(ranked) == np.sum(~invalid)

    # the set of valid items matches
    assert set(ids[~invalid]) == set(ranked.ids())

    # the scores match
    rank_s = ranked.scores("pandas", index="ids")
    assert rank_s is not None
    src_s = items.scores("pandas", index="ids")
    assert src_s is not None

    # make sure the scores were preserved properly
    rank_s, src_s = rank_s.align(src_s, "left")
    assert not np.any(np.isnan(src_s))
    assert np.all(rank_s == src_s)

    # make sure it's sorted
    assert np.all(rank_s.diff()) >= 0


@given(st.integers(min_value=1, max_value=100), scored_lists())
def test_configured_truncation(n, items: ItemList):
    topn = TopNRanker(n=n)
    ranked = topn(items=items)

    ids = items.ids()
    scores = items.scores("numpy")
    assert scores is not None
    invalid = np.isnan(scores)
    _log.info("top %d of %d items, %d invalid", n, len(ids), np.sum(invalid))

    val_items = items[~invalid]

    assert isinstance(ranked, ItemList)
    assert ranked.ordered
    assert len(ranked) == min(n, len(val_items))

    # the scores match
    rank_s = ranked.scores("pandas", index="ids")
    assert rank_s is not None
    src_s = items.scores("pandas", index="ids")
    assert src_s is not None
    src_s = src_s[src_s.notna()]

    # make sure the scores were preserved properly
    rank_s, src_s = rank_s.align(src_s, "left")
    assert not np.any(np.isnan(src_s))
    assert np.all(rank_s == src_s)

    # make sure it's sorted
    assert np.all(rank_s.diff()) >= 0

    # make sure it's the largest
    omitted = ~np.isin(items.ids(), ranked.ids())
    if np.any(omitted) and np.any(~np.isnan(scores[omitted])):
        assert np.all(rank_s >= np.nanmax(scores[omitted]))


@given(st.integers(min_value=1, max_value=100), scored_lists())
def test_runtime_truncation(n, items: ItemList):
    topn = TopNRanker()
    ranked = topn(items=items, n=n)

    ids = items.ids()
    scores = items.scores("numpy")
    assert scores is not None
    invalid = np.isnan(scores)
    _log.info("top %d of %d items, %d invalid", n, len(ids), np.sum(invalid))

    val_items = items[~invalid]

    assert isinstance(ranked, ItemList)
    assert ranked.ordered
    assert len(ranked) == min(n, len(val_items))

    # the scores match
    rank_s = ranked.scores("pandas", index="ids")
    assert rank_s is not None
    src_s = items.scores("pandas", index="ids")
    assert src_s is not None
    src_s = src_s[src_s.notna()]

    # make sure the scores were preserved properly
    rank_s, src_s = rank_s.align(src_s, "left")
    assert not np.any(np.isnan(src_s))
    assert np.all(rank_s == src_s)

    # make sure it's sorted
    assert np.all(rank_s.diff()) >= 0

    # make sure it's the largest
    omitted = ~np.isin(items.ids(), ranked.ids())
    if np.any(omitted) and np.any(~np.isnan(scores[omitted])):
        assert np.all(rank_s >= np.nanmax(scores[omitted]))



#!/usr/bin/env python3
# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Manipulate LensKit versions.

Usage:
    version-tool.py [-v | -vv] [-q]
    version-tool.py [-v | -vv] --github
    version-tool.py [-v | -vv] --run PROGRAM [ARGS...]

Options:
    -v, --verbose
        Ouput verbose logging.
    -q, --quiet
        Print version only, without other text.
    --github
        Write output to GitHub Actions output.
    --run
        Run a program with an updated version.
"""

import logging
import os
import shutil
import subprocess
import sys
from pathlib import Path

import tomlkit
from docopt import docopt

from lenskit._version import lenskit_version

_log = logging.getLogger("lenskit.version-tool")

options = docopt(__doc__ or "", options_first=True)

# set up logging
verbosity = options["--verbose"]
if verbosity > 1:
    level = logging.DEBUG
elif verbosity:
    level = logging.INFO
else:
    level = logging.WARNING
logging.basicConfig(level=level, stream=sys.stderr)

# get the version
version = lenskit_version()

if options["--github"]:
    gh_file = os.environ["GITHUB_OUTPUT"]
    with open(gh_file, "at") as ghf:
        print(f"version={version}", file=ghf)

elif options["--run"]:
    ppt_file = Path("pyproject.toml")
    saved = ppt_file.with_suffix(".toml.saved")
    shutil.copy(ppt_file, saved)
    with ppt_file.open("rt") as tf:
        meta = tomlkit.load(tf)

    proj = meta["project"]
    proj["dynamic"].remove("version")
    proj["version"] = str(version)

    try:
        with ppt_file.open("wt") as tf:
            _log.debug("writing updated pyproject.toml")
            tomlkit.dump(meta, tf)

        subprocess.check_call([options["PROGRAM"]] + options["ARGS"])
    finally:
        _log.info("restoring pyproject.toml")
        os.replace(saved, ppt_file)

else:
    # we only want to print the version
    if options["--quiet"]:
        print(version)
    else:
        print("LensKit version", version)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging

import numpy as np

from lenskit.data.items import ItemList
from lenskit.pipeline import Component
from lenskit.pipeline.types import Lazy

_logger = logging.getLogger(__name__)


class FallbackScorer(Component[ItemList]):
    """
    Scoring component that fills in missing scores using a fallback.

    Stability:
        Caller
    """

    config: None

    def __call__(self, primary: ItemList, backup: Lazy[ItemList]) -> ItemList:
        s = primary.scores()
        if s is None:
            return backup.get()

        s = s.copy()
        missing = np.isnan(s)
        if not np.any(missing):
            return primary

        bs = backup.get().scores("pandas", index="ids")
        if bs is None:
            return primary

        s[missing] = bs.reindex(primary.ids()[missing]).values
        return ItemList(primary, scores=s)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

from pytest import mark

from lenskit.data import from_interactions_df
from lenskit.splitting import sample_records
from lenskit.tuning import PipelineTuner, TuningSpec


@mark.slow
@mark.realdata
def test_tune_als(ml_100k, tmpdir):
    spec = TuningSpec.load(Path("pipelines/als-implicit-search.toml"))
    spec.search.method = "random"
    spec.search.max_points = 10
    spec.search.max_epochs = 5

    tuner = PipelineTuner(spec, Path(tmpdir))
    split = sample_records(from_interactions_df(ml_100k), 20000)

    tuner.set_data(split.train, split.test)

    tuner.run()
    print(tuner.best_result())



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import datetime as dt
import re
from pathlib import Path

import click

from lenskit.data.dataset import Dataset
from lenskit.logging import get_logger
from lenskit.splitting import sample_records, split_global_time, split_temporal_fraction

_log = get_logger(__name__)


@click.command("split")
@click.option("--date", type=str, metavar="DATE", help="split at DATE")
@click.option(
    "--fraction", type=float, metavar="FRAC", help="use FRAC fraction interactions for testing"
)
@click.option("--random", is_flag=True, help="split randomly instead of temporally")
@click.option(
    "--min-train-interactions",
    type=int,
    metavar="N",
    help="drop test users without at least N training interactions",
)
@click.option("-d", "--out-dir", type=Path, help="directory to store split data")
@click.argument("data", required=True, type=Path)
def split(
    data: Path,
    out_dir: Path | None,
    date: str | None,
    fraction: float | None,
    random: bool,
    min_train_interactions: int | None,
):
    """
    Prepare a train-test split of a dataset.

    Note: this command does not expose all of LensKit's data-splitting
    capabilities. Use the `lenskit.splitting` API for more fine-grained control.
    """

    ds = Dataset.load(data)
    _log.info("splitting dataset %s", ds.name)

    if date:
        if re.match(r"^\d+(\.\d+)?$", date):
            split_time = dt.datetime.fromtimestamp(float(date))
        elif re.match(r"^\d+-\d+-\d+([T ]\d+:\d+:\d+(\.\d+)?)?", date):
            split_time = dt.datetime.fromisoformat(date)
        else:  # pragma: nocover
            _log.error("invalid date %s", date)
            raise click.UsageError("invalid date")

        split = split_global_time(ds, split_time, filter_test_users=min_train_interactions)

    elif fraction:
        if random:
            n = int(ds.interaction_count * fraction)
            split = sample_records(ds, n)
        else:
            split = split_temporal_fraction(ds, fraction, filter_test_users=min_train_interactions)

    else:  # pragma: nocover
        _log.error("no split method specified")
        raise click.UsageError("must specify a split")

    if out_dir is None:
        train_path = data.with_name(data.name + ".train")
        test_path = data.with_name(data.name + ".test.parquet")
    else:
        train_path = out_dir / "train"
        test_path = out_dir / "test.parquet"

    split.train.save(train_path)
    split.test.save_parquet(test_path)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

from pytest import approx

from lenskit.data import Dataset, ItemList
from lenskit.metrics.ranking import MeanPopRank
from lenskit.testing import DemoRecs, demo_recs  # noqa: F401


def _test_mpr(ml_ds, items, rel, **k):
    recs = ItemList(items, ordered=True)
    truth = ItemList(rel)
    return MeanPopRank(ml_ds, **k).measure_list(recs, truth)


def test_mpr_empty_none(ml_ds: Dataset):
    mpr = _test_mpr(ml_ds, [], [1, 3])
    assert np.isnan(mpr)


def test_mpr_label(ml_ds: Dataset):
    mpr = MeanPopRank(ml_ds)
    assert mpr.label == "MeanPopRank"

    mpr = MeanPopRank(ml_ds, k=10)
    assert mpr.label == "MeanPopRank@10"


def test_mpr_single_max(ml_ds: Dataset):
    items = ml_ds.item_stats().sort_values("count", ascending=False)
    mpr = _test_mpr(ml_ds, [items.index[0]], [1, 3])
    assert mpr == approx(1.0)


def test_mpr_single_min(ml_ds: Dataset):
    items = ml_ds.item_stats().sort_values("count", ascending=False)
    mpr = _test_mpr(ml_ds, [items.index[-1]], [1, 3])
    assert mpr == approx(0.0)


def test_mpr_min_max(ml_ds: Dataset):
    items = ml_ds.item_stats().sort_values("count", ascending=False)
    mpr = _test_mpr(ml_ds, [items.index[-1], items.index[0]], [1, 3])
    assert mpr == approx(0.5)


def test_mpr_random(ml_ds: Dataset, rng: np.random.Generator):
    items = rng.choice(ml_ds.items.ids(), 100, replace=False)
    mpr = _test_mpr(ml_ds, items, [])

    stats = ml_ds.item_stats()
    counts = stats["count"]
    counts = counts[counts > 0]
    ranks = counts.rank(method="average", ascending=True) / len(counts)
    assert mpr == approx(ranks.reindex(items, fill_value=0).mean(), rel=1.0e-3)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Simple (non-iterative) evaluation of points.
"""

from __future__ import annotations

import pickle
from pathlib import Path

import ray
import ray.tune
import ray.tune.result
import torch
from pydantic_core import to_json
from structlog.stdlib import BoundLogger

from lenskit.batch import BatchPipelineRunner
from lenskit.logging import Task, get_logger
from lenskit.logging.worker import send_task
from lenskit.pipeline import Component, Pipeline, PipelineBuilder
from lenskit.pipeline.components import Placeholder
from lenskit.pipeline.nodes import ComponentConstructorNode, ComponentInstanceNode
from lenskit.random import make_seed
from lenskit.splitting.split import TTSplit
from lenskit.state import ParameterContainer
from lenskit.training import ModelTrainer, TrainingOptions, UsesTrainer

from ._job import TuningJobData
from ._measure import measure_pipeline

_log = get_logger(__name__)


class IterativeEval(ray.tune.Trainable):
    """
    A simple hyperparameter point evaluator using non-iterative model training.
    """

    job: TuningJobData
    log: BoundLogger
    data: TTSplit
    task: Task

    pipeline: Pipeline
    trainer: ModelTrainer

    def setup(self, config, job):
        self.job = job
        self.data = self.job.data
        name = self.job.pipeline.meta.name

        self.log = _log.bind(pipeline=name, dataset=self.job.data_name)

        self.task = Task(
            label=f"tune {name}",
            tags=["tune"],
            reset_hwm=True,
            subprocess=True,
        )

        comp_name = self.job.spec.component_name
        assert comp_name is not None
        self.log.info("setting up base pipeline")
        pb = PipelineBuilder.from_config(self.job.pipeline)
        comp_node = pb.node(comp_name)
        assert isinstance(comp_node, ComponentConstructorNode)
        assert isinstance(comp_node.constructor, type)

        pb.replace_component(comp_name, Placeholder)
        wrapper_pipe = pb.build()

        self.log.info("pre-training pipeline")
        wrapper_pipe.train(self.data.train)

        self.log.info("adding untrained scorer", config=config)
        cfg_pipe = self.job.pipeline.merge_component_configs({comp_name: config})
        pb = PipelineBuilder.from_pipeline(wrapper_pipe)
        pb.replace_component(
            comp_name, comp_node.constructor, cfg_pipe.components[comp_name].config
        )
        self.pipeline = pb.build()
        comp_node = self.pipeline.node(comp_name)

        assert isinstance(comp_node, ComponentInstanceNode)
        self.scorer = comp_node.component

        assert isinstance(self.scorer, Component)
        assert isinstance(self.scorer, UsesTrainer)

        send_task(self.task)

        self.log.info("creating model trainer", config=self.scorer.config)
        options = TrainingOptions(
            rng=make_seed(self.job.random_seed, to_json(self.scorer.dump_config()))
        )
        self.trainer = self.scorer.create_trainer(self.data.train, options)
        send_task(self.task)

        self.runner = BatchPipelineRunner(n_jobs=1)  # single-threaded inside tuning
        self.runner.recommend()
        if self.pipeline.node("rating-predictor", missing=None) is not None:
            self.runner.predict()

    def step(self):
        epoch = self.iteration
        epoch_limit = self.job.spec.search.max_epochs
        if epoch > epoch_limit:
            return {ray.tune.result.DONE: True}

        elog = self.log.bind(epoch=epoch)
        with Task(f"epoch {self.iteration}", tags=["tuning", "epoch"]) as e_task:
            elog.debug("beginning training epoch")
            with Task(
                f"training epoch {self.iteration}", tags=["tuning", "epoch", "train"]
            ) as t_task:
                vals = self.trainer.train_epoch()
            elog.debug("epoch training finished", result=vals, duration=t_task.friendly_duration)

            elog.debug("generating recommendations", n_queries=len(self.data.test))
            with Task(
                f"measuring epoch {self.iteration}", tags=["tuning", "epoch", "measure"]
            ) as m_task:
                metrics = measure_pipeline(self.job.spec, self.pipeline, self.data.test)
                metrics["max_epochs"] = epoch_limit
                if epoch == epoch_limit:
                    metrics[ray.tune.result.DONE] = True

            metrics["epoch_train_s"] = t_task.duration
            metrics["epoch_measure_s"] = m_task.duration
            elog.debug("epoch measurement finished", duration=m_task.friendly_duration)

        send_task(self.task)
        elog.info("epoch complete", duration=e_task.friendly_duration)
        return metrics

    def cleanup(self):
        self.trainer.finalize()

    def save_checkpoint(self, checkpoint_dir: str):
        cpdir = Path(checkpoint_dir)
        log = self.log.bind(epochs=self.iteration)
        if isinstance(self.trainer, ParameterContainer):
            log.info("saving checkpoint")
            torch.save(
                self.trainer.get_parameters(),
                cpdir / "model.pt",
                pickle_protocol=pickle.HIGHEST_PROTOCOL,
            )
        else:
            log.warning("trainer does not implement ParameterContainer, pickling")
            with open(cpdir / "model.pkl", "wb") as pkf:
                pickle.dump(self.trainer, pkf)

    def load_checkpoint(self, checkpoint_dir: str):
        cpdir = Path(checkpoint_dir)

        self.log.info("resuming from checkpoint", epochs=self.iteration)

        ptf = cpdir / "model.pt"
        if ptf.exists():
            assert isinstance(self.trainer, ParameterContainer)
            self.trainer.load_parameters(torch.load(ptf, weights_only=False))
        else:
            with open(cpdir / "model.pkl", "rb") as pkf:
                self.trainer = pickle.load(pkf)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Mixins with commonly-used component configuration capabilities.
"""

from __future__ import annotations

from pydantic import PositiveInt, model_validator


class EmbeddingSizeMixin:
    """
    Mixin for configuring embedding sizes (# of latent dimensions).

    Component configuration classes can extend this class to inherit a
    standardized definition of an embedding size, along with useful behavior
    like configuring with base-2 logs.

    Example usage:

    .. code:: python

        class SVDConfig(EmbeddingSizeMixin, BaseModel):
            pass

        cfg = SVDConfig(embedding_size=32)
    """

    embedding_size: PositiveInt
    """
    The dimension of user and item embeddings (number of latent features to
    learn).
    """

    @model_validator(mode="before")
    @classmethod
    def lkmv_embedding_size(cls, data):
        match data:
            case {"embedding_size_exp": e}:
                # convert embedding size and remove from data
                return {"embedding_size": 2**e} | {
                    k: v
                    for k, v in data.items()
                    if k not in ("embedding_size_exp", "embedding_size")
                }
            case _:
                return data



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Pydantic models for pipeline configuration and serialization support.
"""

# pyright: strict
from __future__ import annotations

import base64
import pickle
import warnings
from collections import OrderedDict
from hashlib import sha256
from types import FunctionType
from typing import Annotated, Literal, Mapping

from annotated_types import Predicate
from deepmerge import always_merger
from pydantic import AliasChoices, BaseModel, Field, JsonValue, TypeAdapter, ValidationError
from typing_extensions import Any, Self

from lenskit.diagnostics import PipelineError

from ._hooks import HookEntry
from .components import Component
from .nodes import ComponentConstructorNode, ComponentInstanceNode, ComponentNode, InputNode
from .types import type_string


class PipelineHook(BaseModel):
    """
    A single entry in a pipeline's hook configuration.
    """

    function: str
    priority: Annotated[int, Predicate(lambda x: x != 0)] = 1

    @classmethod
    def from_entry(cls, hook: HookEntry[Any]):
        if not isinstance(hook, FunctionType):  # type: ignore
            warnings.warn(f"hook {hook.function} is not a function")
        function = f"{hook.function.__module__}:{hook.function.__qualname__}"
        return cls(function=function, priority=hook.priority)


class PipelineHooks(BaseModel):
    """
    Hook specifications for a pipeline.
    """

    run: dict[str, list[PipelineHook]] = {}


class PipelineOptions(BaseModel, extra="allow"):
    """
    Options used for pipeline assembly, particularly for extending pipelines.
    """

    base: str | None = None
    "Apply this configuration to a base pipeline."

    default_length: int | None = Field(
        default=None, validation_alias=AliasChoices("default_length", "n")
    )
    """
    For top-*N* base pipelines, the default ranking length to configure.
    """

    fallback_predictor: bool | None = None
    """
    For rating prediction pipelines, enable or disable the default fallback
    predictor.
    """


class PipelineConfig(BaseModel):
    """
    Root type for serialized pipeline configuration.  A pipeline config contains
    the full configuration, components, and wiring for the pipeline, but does
    not contain the learned parameters.

    Stability:
        Full
    """

    options: PipelineOptions | None = None
    "Options for assembling the final pipeline."
    meta: PipelineMeta
    "Pipeline metadata."
    inputs: list[PipelineInput] = []
    "Pipeline inputs."
    components: OrderedDict[str, PipelineComponent] = Field(default_factory=OrderedDict)
    "Pipeline components, with their configurations and wiring."
    aliases: dict[str, str] = {}
    "Pipeline node aliases."
    default: str | None = None
    "The default node for running this pipeline."
    literals: dict[str, PipelineLiteral] = {}
    "Literals"
    hooks: PipelineHooks = PipelineHooks()
    "The hooks configured for the pipeline."

    def merge_component_configs(
        self, configs: Mapping[str, Mapping[str, JsonValue]]
    ) -> PipelineConfig:
        """
        Merge component configuration options into the pipeline configuration.

        This returns a modified copy of the pipeline with the applied
        configurations, and does not modify the configuration in-place.
        """
        pipe = self.model_copy(deep=True)
        for name, cfg in configs.items():
            try:
                comp = pipe.components[name]
            except KeyError:
                raise PipelineError(f"unknown component {name}")

            comp.config = always_merger.merge(comp.config or {}, cfg)

        return pipe


class PipelineMeta(BaseModel):
    """
    Pipeline metadata.

    Stability:
        Full
    """

    name: str | None = None
    "The pipeline name."
    version: str | None = None
    "The pipeline version."
    hash: str | None = None
    """
    The pipeline configuration hash.  This is optional, particularly when
    hand-crafting pipeline configuration files.
    """


class PipelineInput(BaseModel):
    """
    Spcification of a pipeline input.

    Stability:
        Full
    """

    name: str
    "The name for this input."
    types: set[str] | None
    "The list of types for this input."

    @classmethod
    def from_node(cls, node: InputNode[Any]) -> Self:
        if node.types is not None:
            types = {type_string(t) for t in node.types}
        else:
            types = None

        return cls(name=node.name, types=types)


class PipelineComponent(BaseModel):
    """
    Specification of a pipeline component.
    """

    code: str = Field(validation_alias=AliasChoices("code", "class"))
    """
    The path to the component's implementation, either a class or a function.
    This is a Python qualified path of the form ``module:name``.
    """

    config: Mapping[str, JsonValue] | None = None
    """
    The component configuration.  If not provided, the component will be created
    with its default constructor parameters.
    """

    inputs: dict[str, str] = {}
    """
    The component's input wirings, mapping input names to node names.  For
    certain meta-nodes, it is specified as a list instead of a dict.
    """

    @classmethod
    def from_node(cls, node: ComponentNode[Any]) -> Self:
        match node:
            case ComponentInstanceNode(_name, comp):
                config = None
                if isinstance(comp, FunctionType):
                    ctype = comp
                else:
                    ctype = comp.__class__
                    if isinstance(comp, Component):
                        config = comp.dump_config()
            case ComponentConstructorNode(_name, ctype, config):
                if isinstance(ctype, type) and issubclass(ctype, Component):
                    config = TypeAdapter[Any](ctype.config_class()).dump_python(config, mode="json")
                else:
                    config = None
            case _:
                raise TypeError("unexpected node type")

        code = f"{ctype.__module__}:{ctype.__qualname__}"

        return cls(code=code, config=config)


class PipelineLiteral(BaseModel):
    """
    Literal nodes represented in the pipeline.

    Stability:
        Full
    """

    encoding: Literal["json", "base85"]
    value: JsonValue

    @classmethod
    def represent(cls, data: Any) -> Self:
        try:
            return cls(encoding="json", value=data)
        except ValidationError:
            # data is not basic JSON values, so let's pickle it
            dbytes = pickle.dumps(data)
            return cls(encoding="base85", value=base64.b85encode(dbytes).decode("ascii"))

    def decode(self) -> Any:
        "Decode the represented literal."
        match self.encoding:
            case "json":
                return self.value
            case "base85":
                assert isinstance(self.value, str)
                return pickle.loads(base64.b85decode(self.value))


def hash_config(config: BaseModel) -> str:
    """
    Compute the hash of a configuration model.

    Stability:
        Internal
    """
    json = config.model_dump_json(exclude_none=True)
    h = sha256()
    h.update(json.encode())
    return h.hexdigest()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Sequence

from docutils import nodes
from docutils.parsers.rst.directives.admonitions import Admonition
from sphinx.addnodes import pending_xref
from sphinx.application import Sphinx
from sphinx.util.typing import ExtensionMetadata

STABILITY_LEVELS = {
    "full": """
This API is at the **full** stability level; breaking changes for both callers
and implementers will be reserved for annual major version bumps. See
:ref:`stability-levels` for details.
""",
    "caller": """
This API is at the **caller** stability level: breaking changes for code calling
this function or class will be reserved for annual major version bumps, but minor
versions may introduce changes that break subclasses or reimplementations. See
:ref:`stability-levels` for details.
""",
    "testing": """
This API is at the **testing**: we will avoid gratuituous breaking changes for
callers, but may make such changes in minor versions with clear statements in the
release notes.  No stability guarantees are made for subclasses or re-implementers.
See :ref:`stability-levels` for details.
""",
    "internal": """
This API is at the **internal** or **experimental** stability level: it may
change at any time, and breaking changes will not necessarily be described in
the release notes. See :ref:`stability-levels` for details.
""",
}


class StabilityDirective(Admonition):
    """
    A directive to note an API's stability.
    """

    required_arguments = 1

    def run(self) -> Sequence[nodes.Node]:
        level = self.arguments[0]
        text = STABILITY_LEVELS.get(level, "")
        self.arguments = ["Stability: " + level.capitalize()]
        self.options["class"] = ["note"]
        self.content.insert(0, text, "unknown")
        return super().run()


def scan_stability_notes(app, domain, objtype, contentnode):
    for node in contentnode.findall(_is_stability_field):
        vals = node.traverse(nodes.field_body)
        if not vals:
            continue

        body: nodes.field_body = vals[0]
        text = body.astext()
        if text.lower() in STABILITY_LEVELS:
            # we found a stability level — add reference
            body.clear()
            body.extend(
                [
                    nodes.Text(text.capitalize()),
                    nodes.Text(" (see "),
                    nodes.inline(
                        "",
                        "",
                        pending_xref(
                            "",
                            nodes.inline(
                                "",
                                "stability-levels",
                                classes=["xref", "std", "std-ref"],
                            ),
                            refdoc="api/operations",
                            refdomain="std",
                            reftype="ref",
                            reftarget="stability-levels",
                            refexplicit=False,
                            refwarn=True,
                        ),
                    ),
                    nodes.Text(")."),
                ]
            )


def _is_stability_field(node):
    if isinstance(node, nodes.field):
        names = node.traverse(nodes.field_name)
        if names:
            return names[0].astext() == "Stability"


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_directive("stability", StabilityDirective)
    app.connect("object-description-transform", scan_stability_notes)

    return {
        "version": "2025.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Generate summaries of datasets.
"""

from os import PathLike
from pathlib import Path
from typing import TextIO

import pyarrow as pa
import pyarrow.compute as pc
from humanize import metric, naturalsize
from prettytable import PrettyTable, TableStyle
from structlog.stdlib import BoundLogger

from lenskit.data.matrix import SparseRowType
from lenskit.logging import get_logger

from .container import DataContainer
from .dataset import Dataset
from .schema import AttrLayout, ColumnSpec

_log = get_logger(__name__)


def save_stats(data: Dataset | DataContainer, out: str | Path | PathLike[str] | TextIO):
    """
    Save dataset statistics to a file or output stream.
    """
    if isinstance(data, Dataset):
        data = data._data

    log = _log.bind()

    if hasattr(out, "write"):
        _write_stats(data, out, log)  # type: ignore
    else:
        log = log.bind(file=str(out))
        with open(out, "wt") as stats:  # type: ignore
            _write_stats(data, stats, log)


def _write_stats(data: DataContainer, out: TextIO, log: BoundLogger):
    if data.schema.name:
        print("# Summary of", data.schema.name, file=out)
    else:
        print("# Summary of Unnamed Data", file=out)
    print(file=out)

    _write_entities(data, out, log)
    _write_relationships(data, out, log)

    print("## Data Tables\n", file=out)
    print(table_stats(data), file=out)


def _write_entities(data: DataContainer, out: TextIO, log: BoundLogger):
    print("## Entities\n", file=out)
    log.debug("summarizing entities")
    for name, schema in data.schema.entities.items():
        print(
            "- {} ({}, {} attributes)".format(
                name, metric(data.tables[name].num_rows), len(schema.attributes)
            ),
            file=out,
        )
    print(file=out)

    for name in data.schema.entities.keys():
        _write_entity_info(data, name, out, log)


def _write_entity_info(data: DataContainer, name: str, out: TextIO, log: BoundLogger):
    tbl = data.tables[name]
    log = log.bind(entity=name)
    log.debug("describing entity")
    print(f"### Entity `{name}`\n", file=out)
    print("- {:,d} instances\n".format(tbl.num_rows), file=out)

    attributes = data.schema.entities[name].attributes
    if attributes:
        print("#### Attributes\n", file=out)
        print(_attr_table(tbl, attributes), file=out)
        print(file=out)

    print("#### Schema\n", file=out)
    print("```", file=out)
    print(tbl.schema.to_string(), file=out)
    print("```\n", file=out)


def _write_relationships(data: DataContainer, out: TextIO, log: BoundLogger):
    print("## Relationships\n", file=out)
    log.debug("summarizing relationships")
    for name, schema in data.schema.relationships.items():
        print(
            "- {} ({}, {} attributes{})".format(
                name,
                metric(data.tables[name].num_rows),
                len(schema.attributes),
                ", interaction" if schema.interaction else "",
            ),
            file=out,
        )
    print(file=out)

    for name in data.schema.relationships.keys():
        _write_relationship_info(data, name, out, log)


def _write_relationship_info(data: DataContainer, name: str, out: TextIO, log: BoundLogger):
    tbl = data.tables[name]
    log = log.bind(relationship=name)
    log.debug("describing relationship")
    print(f"### Relationship `{name}`\n", file=out)
    print("- {:,d} records\n".format(tbl.num_rows), file=out)

    print("#### Entities\n", file=out)
    pt = PrettyTable()
    pt.set_style(TableStyle.MARKDOWN)
    pt.field_names = ["Name", "Class", "Unique Count"]
    pt.align["Name"] = "l"
    pt.align["Unique Count"] = "r"
    pt.custom_format["Unique Count"] = lambda _, v: "{:,d}".format(v)
    for e_name, e_cls in data.schema.relationships[name].entities.items():
        e_cls = e_cls or e_name
        e_col = tbl.column(e_name + "_num")
        pt.add_row([e_name, e_cls, pc.count_distinct(e_col).as_py()])
    print(pt, file=out)
    print(file=out)

    attributes = data.schema.relationships[name].attributes
    if attributes:
        print("#### Attributes\n", file=out)
        print(_attr_table(tbl, attributes), file=out)
        print(file=out)


def _attr_table(tbl: pa.Table, attributes: dict[str, ColumnSpec]):
    pt = PrettyTable()
    pt.set_style(TableStyle.MARKDOWN)
    pt.field_names = ["Name", "Layout", "Type", "Dimension", "Count", "Size"]
    pt.align["Name"] = "l"
    pt.align["Dimension"] = "r"
    pt.align["Count"] = "r"
    pt.custom_format["Count"] = lambda _, v: "{:,d}".format(v)
    pt.align["Size"] = "r"
    pt.custom_format["Size"] = lambda _, v: naturalsize(v, binary=True)
    for a_name, a_schema in attributes.items():
        field = tbl.field(a_name)
        col = tbl.column(a_name)
        vtype = field.type
        dim = "-"
        match a_schema.layout:
            case AttrLayout.LIST:
                assert isinstance(vtype, pa.ListType)
                vtype = vtype.value_type
            case AttrLayout.VECTOR:
                assert isinstance(vtype, (pa.ListType, pa.FixedSizeListType))
                vtype = vtype.value_type
                dim = "{:,d}".format(a_schema.vector_size)
            case AttrLayout.SPARSE:
                assert isinstance(vtype, SparseRowType)
                vtype = vtype.value_type
                dim = "{:,d}".format(a_schema.vector_size)

        pt.add_row(
            [a_name, a_schema.layout.value, vtype, dim, tbl.num_rows - col.null_count, col.nbytes]
        )

    return pt


def table_stats(data: DataContainer) -> PrettyTable:
    tbl = PrettyTable()
    tbl.set_style(TableStyle.MARKDOWN)
    tbl.field_names = ["Name", "Rows", "Bytes"]
    tbl.align["Name"] = "l"
    tbl.align["Rows"] = "r"
    tbl.custom_format["Rows"] = lambda _, v: "{:,d}".format(v)
    tbl.align["Bytes"] = "r"
    tbl.custom_format["Bytes"] = lambda _, v: naturalsize(v, binary=True)

    for name, table in data.tables.items():
        tbl.add_row([name, table.num_rows, table.nbytes])

    return tbl



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging

import numpy as np
import pandas as pd

from pytest import importorskip, mark

from lenskit.data import ItemList, from_interactions_df
from lenskit.metrics import quick_measure_model
from lenskit.testing import BasicComponentTests, ScorerTests

nmf = importorskip("lenskit.sklearn.nmf")

_log = logging.getLogger(__name__)

simple_df = pd.DataFrame(
    {"item": [1, 1, 2, 3], "user": [10, 12, 10, 13], "rating": [4.0, 3.0, 5.0, 2.0]}
)
simple_ds = from_interactions_df(simple_df)


class TestNMF(BasicComponentTests, ScorerTests):
    component = nmf.NMFScorer
    config = nmf.NMFConfig(max_iter=10, n_components=25)
    expected_ndcg = 0.22


def test_nmf_basic_build():
    algo = nmf.NMFScorer(n_components=2)
    algo.train(simple_ds)

    assert algo.user_components.shape == (3, 2)


def test_nmf_predict_basic():
    _log.info("NMF input data:\n%s", simple_df)
    algo = nmf.NMFScorer(n_components=2)
    algo.train(simple_ds)
    _log.info("user matrix:\n%s", str(algo.user_components))
    _log.info("item matrix:\n%s", str(algo.item_components))

    preds = algo(10, ItemList([3]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert preds.loc[3] >= 0
    assert preds.loc[3] <= 5


def test_nmf_predict_bad_item():
    algo = nmf.NMFScorer(n_components=2)
    algo.train(simple_ds)

    preds = algo(10, ItemList([4]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 4
    assert np.isnan(preds.loc[4])


def test_nmf_predict_bad_user():
    algo = nmf.NMFScorer(n_components=2)
    algo.train(simple_ds)

    preds = algo(50, ItemList([3]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert np.isnan(preds.loc[3])



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
FunkSVD (biased MF).
"""

import time
from dataclasses import dataclass

import numpy as np
import pyarrow as pa
from pydantic import AliasChoices, BaseModel, Field, NonNegativeFloat, PositiveFloat, PositiveInt
from typing_extensions import override

from lenskit._accel import FunkSVDTrainer
from lenskit.basic import BiasModel, Damping
from lenskit.config.common import EmbeddingSizeMixin
from lenskit.data import Dataset, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.data.types import NPMatrix
from lenskit.logging import Stopwatch, get_logger
from lenskit.logging.progress._dispatch import item_progress
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_logger = get_logger(__name__)

INITIAL_VALUE = 0.1


class FunkSVDConfig(EmbeddingSizeMixin, BaseModel):
    "Configuration for :class:`FunkSVDScorer`."

    embedding_size: PositiveInt = Field(
        default=50, validation_alias=AliasChoices("embedding_size", "features")
    )
    """
    Number of latent features.
    """
    epochs: PositiveInt = 100
    """
    Number of training epochs (per feature).
    """
    learning_rate: PositiveFloat = 0.001
    """
    Gradient descent learning rate.
    """
    regularization: NonNegativeFloat = 0.015
    """
    Parameter regularization.
    """
    damping: Damping = 5.0
    """
    Bias damping term.
    """
    range: tuple[float, float] | None = None
    """
    Min/max range of ratings to clamp output.
    """


@dataclass
class FunkSVDTrainingParams:
    learning_rate: float
    regularization: float
    rating_min: float
    rating_max: float


@dataclass
class FunkSVDTrainingData:
    users: pa.Int32Array
    items: pa.Int32Array
    ratings: pa.FloatArray


class FunkSVDScorer(Trainable, Component[ItemList]):
    """
    FunkSVD explicit-feedback matrix factoriation.  FunkSVD is a regularized
    biased matrix factorization technique trained with featurewise stochastic
    gradient descent.

    See the base class :class:`.MFPredictor` for documentation on the estimated
    parameters you can extract from a trained model.

    .. deprecated:: LKPY

        This scorer is kept around for historical comparability, but ALS
        :class:`~lenskit.als.BiasedMF` is usually a better option.

    Stability:
        Caller
    """

    config: FunkSVDConfig

    bias: BiasModel
    users: Vocabulary
    user_embeddings: NPMatrix
    items: Vocabulary
    item_embeddings: NPMatrix

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        """
        Train a FunkSVD model.

        Args:
            ratings: the ratings data frame.
        """
        if hasattr(self, "item_embeddings") and not options.retrain:
            return

        log = _logger.bind(dataset=data.name)

        timer = Stopwatch()
        rmat = data.interactions().matrix().scipy(attribute="rating", layout="coo")

        n_users, n_items = rmat.shape
        n_ratings = rmat.nnz
        log = log.bind(n_users=n_users, n_items=n_items, n_ratings=n_ratings)

        log.info("[%s] fitting bias model", timer)
        self.bias = BiasModel.learn(data, damping=self.config.damping)

        log.info("[%s] preparing rating data", timer)
        log.debug("shuffling rating data")
        shuf = np.arange(n_ratings, dtype=np.int_)
        rng = options.random_generator()
        rng.shuffle(shuf)
        users = pa.array(rmat.row[shuf], pa.int32())
        items = pa.array(rmat.col[shuf], pa.int32())
        ratings = pa.array(rmat.data[shuf], pa.float32())

        del rmat  # don't need it now that data is in Arrow

        log.debug("[%s] computing initial estimates", timer)
        est = np.full(n_ratings, self.bias.global_bias, dtype=np.float32)
        if self.bias.item_biases is not None:
            est += self.bias.item_biases[items.to_numpy()]
        if self.bias.user_biases is not None:
            est += self.bias.user_biases[users.to_numpy()]

        log.debug("[%s] initializing embeddings")
        esize = self.config.embedding_size
        uemb = np.full([n_users, esize], INITIAL_VALUE, dtype=np.float32, order="F")
        iemb = np.full([n_items, esize], INITIAL_VALUE, dtype=np.float32, order="F")

        if self.config.range is not None:
            rmin, rmax = self.config.range
        else:
            rmin = -np.inf
            rmax = np.inf

        config = FunkSVDTrainingParams(
            learning_rate=self.config.learning_rate,
            regularization=self.config.regularization,
            rating_min=rmin,
            rating_max=rmax,
        )
        train_data = FunkSVDTrainingData(users=users, items=items, ratings=ratings)

        trainer = FunkSVDTrainer(config, train_data, uemb, iemb)

        log.info("beginning FunkSVD training")
        with item_progress("FunkSVD dimensions", self.config.embedding_size) as pb:
            for f in range(self.config.embedding_size):
                flog = log.bind(dim=f)
                start = time.perf_counter()
                trail = INITIAL_VALUE * INITIAL_VALUE * (esize - f - 1)
                with item_progress(f"Feature {f + 1} epochs", self.config.epochs) as epb:
                    for e in range(self.config.epochs):
                        elog = flog.bind(epoch=e)
                        rmse = trainer.feature_epoch(f, est, trail)
                        elog.debug("[%s] finished epoch with RMSE %.3f", timer, rmse)
                        epb.update()

                end = time.perf_counter()
                flog.info(
                    "[%s] finished feature %d (RMSE=%f) in %.2fs", timer, f, rmse, end - start
                )
                pb.update()

                est = np.clip(
                    est + uemb[users.to_numpy(), f] * iemb[items.to_numpy(), f], rmin, rmax
                )

        _logger.info("finished model training in %s", timer)

        self.users = data.users
        self.items = data.items
        self.user_embeddings = uemb
        self.item_embeddings = iemb

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)

        user_id = query.user_id
        user_num = None
        if user_id is not None:
            user_num = self.users.number(user_id, missing=None)
        if user_num is None:
            _logger.debug("unknown user %s", query.user_id)
            return ItemList(items, scores=np.nan)

        u_feat = self.user_embeddings[user_num, :]

        item_nums = items.numbers(vocabulary=self.items, missing="negative")
        item_mask = item_nums >= 0
        i_feats = self.item_embeddings[item_nums[item_mask], :]

        scores = np.full((len(items),), np.nan, dtype=np.float64)
        scores[item_mask] = i_feats @ u_feat
        biases, _ub = self.bias.compute_for_items(items, user_id, query.user_items)
        scores += biases

        return ItemList(items, scores=scores)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import pandas as pd
from pydantic import JsonValue

from lenskit import Pipeline, predict, recommend
from lenskit.data import GenericKey, ItemList, ItemListCollection
from lenskit.data.collection import key_dict
from lenskit.logging import Task, get_logger, item_progress
from lenskit.metrics import (
    DCG,
    NDCG,
    RBP,
    RMSE,
    Hit,
    RecipRank,
    call_metric,
)

from .spec import TuningSpec

_log = get_logger(__name__)


def measure_pipeline(
    spec: TuningSpec,
    pipe: Pipeline,
    test_users: ItemListCollection,
    train_task: Task | None = None,
    test_task: Task | None = None,
):
    # TODO: integrate this with metric collectors
    metrics = []
    with item_progress("Measuring for users", total=len(test_users)) as pb:
        for key, test in test_users.items():
            metrics.append(measure_request(spec, pipe, key, test))
            pb.update()

    metric_df = pd.DataFrame.from_records(metrics)
    metric_df = metric_df.drop(columns=["user_id"])
    agg_metrics = metric_df.mean().to_dict()

    if train_task is not None:
        agg_metrics["TrainTask"] = train_task.task_id
        agg_metrics["TrainTime"] = train_task.duration
        agg_metrics["TrainCPU"] = train_task.cpu_time
    if test_task is not None:
        test_task.update_resources()
        agg_metrics["TestTask"] = test_task.task_id
        agg_metrics["TestTime"] = test_task.duration
        agg_metrics["TestCPU"] = test_task.cpu_time
    return agg_metrics


def measure_request(spec: TuningSpec, pipe: Pipeline, key: GenericKey, test: ItemList):
    recs = recommend(pipe, query=key.user_id, n=spec.search.list_length)  # type: ignore
    if pipe.node("rating-predictor", missing=None) is not None:
        preds = predict(pipe, query=key.user_id, items=test)  # type: ignore
    else:
        preds = None
    return measure_list(spec, key, recs, preds, test)


def measure_list(
    spec: TuningSpec,
    key: GenericKey,
    recs: ItemList,
    preds: ItemList | None,
    test: ItemList,
):
    log = _log.bind(key=key)
    log.debug("measuring recommendation list")

    metrics: dict[str, JsonValue] = key_dict(key)
    metrics["RBP"] = float(call_metric(RBP, recs, test))
    metrics["DCG"] = float(call_metric(DCG, recs, test))
    metrics["NDCG"] = float(call_metric(NDCG, recs, test))
    metrics["RecipRank"] = float(call_metric(RecipRank, recs, test))
    metrics["Hit10"] = float(call_metric(Hit, recs, test, k=10))

    if preds is not None:
        log.debug("measuring rating predictions")
        metrics["RMSE"] = call_metric(RMSE, preds, test)

    return metrics



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
from typing_extensions import override

from lenskit.data import Dataset, ItemList
from lenskit.logging import get_logger
from lenskit.stats import gini

from ._base import DecomposedMetric, RankingMetricBase
from ._weighting import GeometricRankWeight, RankWeight

_log = get_logger(__name__)


class GiniBase(DecomposedMetric, RankingMetricBase):
    """
    Base class for Gini diversity / popularity concentration metrics.
    """

    item_count: int

    def __init__(
        self,
        *,
        items: int | pd.Series | pd.DataFrame | Dataset,
        k: int | None = None,
    ):
        super().__init__(k)
        if isinstance(items, int):
            self.item_count = items
        elif isinstance(items, Dataset):
            self.item_count = items.item_count
        else:
            self.item_count = len(items)


class ListGini(GiniBase):
    """
    Measure item diversity of recommendations with the Gini coefficient.

    This computes the Gini coefficient of the *number of lists* that each item
    appears in.

    Args:
        k:
            The maximum recommendation list length.
        items:
            The total number of items, a data frame or series of item data, or a
            dataset. If a frame or series is provided, its length will be used
            as the number of items.  If a dataset is provided, its item count
            will be used.

    Stability:
        Caller
    """

    @override
    def compute_list_data(self, output: ItemList, test):
        recs = self.truncate(output)
        return recs.ids(format="arrow")

    @override
    def global_aggregate(self, values: list[pa.Array]):
        log = _log.bind(metric=self.label, item_count=self.item_count)
        log.debug("aggregating for %d lists", len(values))
        chunked = pa.chunked_array(values)
        vc_tbl = pc.value_counts(chunked)
        log.debug("found %d distinct items", len(vc_tbl))
        counts = np.zeros(self.item_count, np.int32)
        counts[: len(vc_tbl)] = np.asarray(vc_tbl.field("counts"), dtype=np.int32)
        return gini(counts)


class ExposureGini(GiniBase):
    """
    Measure exposure distribution of recommendations with the Gini coefficient.

    This uses a weighting model to compute the exposure of each item in each list,
    and computes the Gini coefficient of the total exposure.

    Args:
        k:
            The maximum recommendation list length.
        items:
            The total number of items, a data frame or series of item data, or a
            dataset. If a frame or series is provided, its length will be used
            as the number of items.  If a dataset is provided, its item count
            will be used.
        weight:
            The rank weighting model to use.  Defaults to
            :class:`GeometricRankWeight` with the specified patience parameter.

    Stability:
        Caller
    """

    weight: RankWeight

    def __init__(
        self,
        *,
        items: int | pd.Series | pd.DataFrame | Dataset,
        k: int | None = None,
        weight: RankWeight = GeometricRankWeight(),
    ):
        super().__init__(k=k, items=items)
        self.weight = weight

    @override
    def compute_list_data(self, output: ItemList, test):
        recs = self.truncate(output)
        weights = self.weight.weight(np.arange(1, len(recs) + 1))
        return (recs.ids(format="arrow"), pa.array(weights, type=pa.float32()))

    @override
    def global_aggregate(self, values: list[tuple[pa.Array, pa.FloatArray]]):
        log = _log.bind(metric=self.label, item_count=self.item_count)
        log.debug("aggregating for %d lists", len(values))
        table = pa.Table.from_batches(
            pa.RecordBatch.from_arrays([iv, ev], ["item_id", "exposure"]) for (iv, ev) in values
        )
        exp_tbl = table.group_by("item_id").aggregate([("exposure", "sum")])
        log.debug("found %d distinct items", exp_tbl.num_rows)
        exp = np.zeros(self.item_count, np.float32)
        exp[: exp_tbl.num_rows] = np.asarray(exp_tbl.column("exposure_sum"), dtype=np.float32)
        return gini(exp)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Interfaces and support for model training.
"""

# pyright: strict
from __future__ import annotations

import os
import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass
from time import perf_counter
from typing import TYPE_CHECKING, Literal, Protocol, overload, runtime_checkable

import numpy as np

from lenskit.data.dataset import Dataset
from lenskit.logging import get_logger, item_progress
from lenskit.pipeline.components import Component
from lenskit.random import RNGInput, random_generator

if TYPE_CHECKING:
    import torch

_log = get_logger(__name__)


@dataclass(frozen=True)
class TrainingOptions:
    """
    Options and context settings that govern model training.
    """

    retrain: bool = True
    """
    Whether the model should retrain if it is already trained.  If ``False``,
    the model should cleanly skip training if it is already trained.
    """

    device: str | None = None
    """
    The device on which to train (e.g. ``'cuda'``).  May be ignored if the model
    does not support the specified device.
    """

    rng: RNGInput = None
    """
    Random number generator to use for any randomness in the training process.
    This option contains any `SPEC 7`_-compatible random number generator
    specification; the :func:`~lenskit.random.random_generator` will convert
    that into a NumPy :class:`~numpy.random.Generator`.
    """

    @overload
    def random_generator(self, *, type: Literal["numpy"] = "numpy") -> np.random.Generator: ...
    @overload
    def random_generator(self, *, type: Literal["torch"]) -> torch.Generator: ...
    def random_generator(
        self, *, type: Literal["numpy", "torch"] = "numpy"
    ) -> np.random.Generator | torch.Generator:
        """
        Obtain a random generator from the configured RNG or seed.

        .. note::

            Each call to this method will return a fresh generator from the same
            seed.  Components should call it once at the beginning of their
            training procesess.
        """
        return random_generator(self.rng, type=type)

    def configured_device(self, *, gpu_default: bool = False) -> str:
        """
        Get the configured device, consulting environment variables and defaults
        if necessary.  It looks for a device in the following order:

        1. The :attr:`device`, if specified on this object.
        2. The :envvar:`LK_DEVICE` environment variable.
        3. If CUDA is enabled and ``gpu_default`` is ``True``, return `"cuda"`
        4. The CPU.

        Args:
            gpu_default:
                Whether a CUDA GPU should be preferred if it is available and no
                device has been specified.
        """
        import torch

        if self.device is not None:
            return self.device
        elif dev := os.environ.get("LK_DEVICE", None):
            return dev
        elif gpu_default and torch.cuda.is_available():
            return "cuda"
        else:
            return "cpu"


@runtime_checkable
class Trainable(Protocol):  # pragma: nocover
    """
    Interface for components and objects that can learn parameters from training
    data. It supports training and checking if a component has already been
    trained.  This protocol only captures the concept of trainability; most
    trainable components should have other properties and behaviors as well:

    -   They are usually components (:class:`~lenskit.pipeline.Component`), with
        an appropriate ``__call__`` method.
    -   They should be pickleable.

    ..
        -   They should also usually implement
            :class:`~lenskit.state.ParameterContainer`, to allow the learned
            parameters to be serialized and deserialized without pickling.

    Stability:
        Full
    """

    def train(self, data: Dataset, options: TrainingOptions) -> None:
        """
        Train the model to learn its parameters from a training dataset.

        Args:
            data:
                The training dataset.
            options:
                The training options.
        """
        raise NotImplementedError()


class IterativeTraining(ABC, Trainable):
    """
    Base class for components that support iterative training.  This both
    automates the :meth:`Trainable.train` method for iterative training in terms
    of initialization, epoch, and finalization methods, and exposes those
    methods to client code that may wish to directly control the iterative
    training process.

    .. deprecated: 2025.3

        This base class is deprecated in favor of :class:`ModelTrainer` and
        :class:`UsesTrainer`, which are a better fit for integration with
        tools like Ray Tune.

    Stability:
        Full
    """

    trained_epochs: int = 0
    """
    The number of epochs for which this model has been trained.
    """

    @property
    def expected_training_epochs(self) -> int | None:
        """
        Get the number of training epochs expected to run.  The default
        implementation looks for an ``epochs`` attribute on the configuration
        object (``self.config``).
        """
        cfg = getattr(self, "config", None)
        if cfg:
            return getattr(cfg, "epochs", None)

    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()) -> None:
        """
        Implementation of :meth:`Trainable.train` that uses the training loop.
        It also uses the :attr:`trained_epochs` attribute to detect if the model
        has already been trained for the purposes of honoring
        :attr:`TrainingOptions.retrain`, and updates that attribute as model
        training progresses.
        """
        if self.trained_epochs > 0 and not options.retrain:
            return

        self.trained_epochs = 0
        log = _log.bind(model=f"{self.__class__.__module__}.{self.__class__.__qualname__}")
        log.info("training model")
        n = self.expected_training_epochs
        log.debug("creating training loop")
        loop = self.training_loop(data, options)
        log.debug("beginning training epochs")
        with item_progress("Training epochs", total=n) as pb:
            start = perf_counter()
            for i, metrics in enumerate(loop, 1):
                metrics = metrics or {}
                now = perf_counter()
                elapsed = now - start
                log.info("finished epoch", time="{:.1f}s".format(elapsed), epoch=i, **metrics)
                self.trained_epochs += 1
                start = now
                pb.update()

        log.info("model training finished", epochs=self.trained_epochs)

    @abstractmethod
    def training_loop(
        self, data: Dataset, options: TrainingOptions
    ) -> Iterator[dict[str, float] | None]:
        """
        Training loop implementation, to be supplied by the derived class.  This
        method should return a iterator that, when iterated, will perform each
        training epoch; when training is complete, it should finalize the model
        and signal iteration completion.

        Each epoch can yield metrics, such as training or validation loss, to be
        logged with structured logging and can be used by calling code to do
        other analysis.

        See :ref:`iterative-training` for more details on writing iterative
        training loops.
        """
        raise NotImplementedError()


class UsesTrainer(IterativeTraining, Component, ABC):
    """
    Base class for models that implement :class:`Trainable` via a
    :class:`ModelTrainer`.  This class implements :class:`IterativeTraining` for
    compatibility, but the :class:`IterativeTraining` interface is deprecated.

    The component's configuration must have an ``epochs`` attribute defining the
    number of epochs to train.

    Stability:
        Full
    """

    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()) -> None:
        """
        Implementation of :meth:`Trainable.train` that uses the model trainer.
        """
        if self.trained_epochs > 0 and not options.retrain:
            return

        self.trained_epochs = 0

        log = _log.bind(model=f"{self.__class__.__module__}.{self.__class__.__qualname__}")
        log.info("training model")
        n = self.expected_training_epochs
        assert n is not None, "no training epochs configured"
        log.debug("creating model trainer")
        trainer = self.create_trainer(data, options)

        log.debug("beginning training epochs")
        with item_progress("Training epochs", total=n) as pb:
            start = perf_counter()
            for i in range(1, n + 1):
                metrics = trainer.train_epoch()
                metrics = metrics or {}
                now = perf_counter()
                elapsed = now - start
                log.info("finished epoch", time="{:.1f}s".format(elapsed), epoch=i, **metrics)
                self.trained_epochs += 1
                start = now
                pb.update()

        log.debug("finalizing model")
        trainer.finalize()

        log.info("model training finished", epochs=self.trained_epochs)

    def training_loop(self, data: Dataset, options: TrainingOptions):
        warnings.warn("IteratativeTraining API is deprecated", DeprecationWarning)

        log = _log.bind(model=f"{self.__class__.__module__}.{self.__class__.__qualname__}")
        log.info("training model")
        n = self.expected_training_epochs
        assert n is not None, "no training epochs configured"
        log.debug("creating model trainer")
        trainer = self.create_trainer(data, options)

        log.debug("beginning training epochs")
        with item_progress("Training epochs", total=n) as pb:
            start = perf_counter()
            for i in range(1, n + 1):
                metrics = trainer.train_epoch()
                metrics = metrics or {}
                now = perf_counter()
                elapsed = now - start
                log.info("finished epoch", time="{:.1f}s".format(elapsed), epoch=i, **metrics)
                start = now
                self.trained_epochs += 1
                pb.update()
                yield metrics

        log.debug("finalizing model")
        trainer.finalize()

    @abstractmethod
    def create_trainer(
        self, data: Dataset, options: TrainingOptions
    ) -> ModelTrainer:  # pragma: nocover
        """
        Create a model trainer to train this model.
        """


class ModelTrainer(ABC):
    """
    Protocol implemented by iterative trainers for models.  Models that
    implement :class:`UsesTrainer` will return an object implementing this
    protocol from their :meth:`~UsesTrainer.create_trainer` method.

    ..

        This protocol only defines the core aspects of training a model. Trainers
        should also implement :class:`~lenskit.state.ParameterContainer` to allow
        training to be checkpointed and resumed.

        It is also a good idea for the trainer to be pickleable, but the parameter
        container interface is the primary mechanism for checkpointing.

    Stability:
        Full
    """

    @abstractmethod
    def train_epoch(self) -> dict[str, float] | None:
        """
        Perform one epoch of the training process, optionally returning metrics
        on the training behavior.  After each training iteration, the mmodel
        must be usable.
        """

    def finalize(self) -> None:
        """
        Finish the training process, cleaning up any unneeded data structures
        and doing any finalization steps to the model.

        The default implementation does nothing.
        """
        pass



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from contextvars import ContextVar
from typing import Any

from lenskit.logging import get_logger
from lenskit.pipeline import PipelineBuilder
from lenskit.pipeline.nodes import ComponentInstanceNode
from lenskit.pipeline.types import Lazy, T

_log = get_logger(__name__)
hook_calls = ContextVar(f"lk-{__name__}-hook-calls", default=[])


def proc_hello(msg: str) -> str:
    return f"Hello, {msg}"


def proc_prefix(msg: str, prefix: str) -> str:
    return prefix + msg


def lazy_prefix(msg: Lazy[str], prefix: str, extra: Lazy[str]) -> str:
    return prefix + msg.get()


def _input_hook(
    node: ComponentInstanceNode[Any], input_name: str, input_type: Any, value: Any, **context
) -> Any:
    cs = hook_calls.get()
    _log.debug("input hook called", n=len(cs), msg=value)
    cs.append((node.name, input_name))
    return value


def test_component_input_called():
    build = PipelineBuilder()
    msg = build.create_input("message", str)
    hello = build.add_component("hello", proc_hello, msg=msg)

    build.add_run_hook("component-input", _input_hook)

    pipe = build.build()

    assert len(pipe.config.hooks.run["component-input"]) == 1
    assert pipe.config.hooks.run["component-input"][0].function.endswith(
        "tests.pipeline.test_component_input_hook:_input_hook"
    )
    assert pipe.config.hooks.run["component-input"][0].priority == 1

    hook_calls.set([])
    out = pipe.run(hello, message="world")
    assert out == "Hello, world"

    assert len(hook_calls.get()) == 1


def test_hook_passed_to_clone():
    build = PipelineBuilder()
    msg = build.create_input("message", str)
    _hello = build.add_component("hello", proc_hello, msg=msg)

    build.add_run_hook("component-input", _input_hook)

    pipe = build.build()

    b2 = PipelineBuilder.from_pipeline(pipe)
    cfg2 = b2.build_config()

    assert len(cfg2.hooks.run["component-input"]) == 1
    assert cfg2.hooks.run["component-input"][0].function.endswith(
        "tests.pipeline.test_component_input_hook:_input_hook"
    )
    assert cfg2.hooks.run["component-input"][0].priority == 1


def test_hook_loaded_from_config():
    build = PipelineBuilder()
    msg = build.create_input("message", str)
    _hello = build.add_component("hello", proc_hello, msg=msg)

    build.add_run_hook("component-input", _input_hook)

    cfg = build.build_config()
    print(cfg.model_dump_json(indent=2))

    b2 = PipelineBuilder.from_config(cfg)
    cfg2 = b2.build_config()
    print(cfg2.model_dump_json(indent=2))

    assert len(cfg2.hooks.run["component-input"]) == 1
    assert cfg2.hooks.run["component-input"][0].function.endswith(
        "tests.pipeline.test_component_input_hook:_input_hook"
    )
    assert cfg2.hooks.run["component-input"][0].priority == 1


def test_component_input_called_multi():
    build = PipelineBuilder()
    msg = build.create_input("message", str)
    pfx = build.add_component("prefix", proc_prefix, msg=msg, prefix="good ")
    hello = build.add_component("hello", proc_hello, msg=pfx)

    build.add_run_hook("component-input", _input_hook)

    pipe = build.build()

    hook_calls.set([])
    out = pipe.run(hello, message="friend")
    assert out == "Hello, good friend"

    # we have 3 input edges: 2 to proc_prefix, 1 to proc_hello
    cs = hook_calls.get()
    assert len(hook_calls.get()) == 3
    assert cs == [("prefix", "msg"), ("prefix", "prefix"), ("hello", "msg")]


def test_component_lazy_hook():
    build = PipelineBuilder()
    msg = build.create_input("message", str)
    pfx = build.add_component("prefix", lazy_prefix, msg=msg, prefix="good ", extra="NOTHING")

    build.add_run_hook("component-input", _input_hook)

    pipe = build.build()

    hook_calls.set([])
    out = pipe.run(pfx, message="friend")
    assert out == "good friend"

    # we should be called twice: once for msg, once for prefix. extra is not called.
    cs = hook_calls.get()
    assert len(hook_calls.get()) == 2
    assert cs == [
        ("prefix", "prefix"),
        ("prefix", "msg"),
    ]
